from plover.engine import StartingStrokeState


class AddTranslation:
    def __init__(self, engine):
        self._status = None
        self._engine = engine
        self._translator_states = []
        self._strokes, self._translation = (None, None)
        engine.hook_connect("add_translation", self.trigger)

    def _get_state(self):
        return (
            self._engine.translator_state,
            self._engine.starting_stroke_state,
        )

    def _set_state(self, translator_state, starting_stroke_state):
        self._engine.translator_state = translator_state
        self._engine.starting_stroke_state = starting_stroke_state

    def _clear_state(self, undo=False):
        self._engine.clear_translator_state(undo)
        self._engine.starting_stroke_state = StartingStrokeState()

    def _push_state(self):
        self._translator_states.insert(0, self._get_state())

    def _pop_state(self):
        self._engine.clear_translator_state(True)
        self._set_state(*self._translator_states.pop(0))

    @staticmethod
    def _stroke_filter(key, value):
        return value != "{PLOVER:ADD_TRANSLATION}"

    def send_string(self, s):
        self._translation += s

    def send_backspaces(self, b):
        self._translation = self._translation[:-b]

    def trigger(self):
        if self._status is None:
            self._push_state()
            self._clear_state()
            self._engine.add_dictionary_filter(self._stroke_filter)
            self._status = "strokes"
        elif self._status == "strokes":
            self._engine.remove_dictionary_filter(self._stroke_filter)
            state = self._get_state()[0]
            assert state.translations
            if len(state.translations) == 1:
                # Abort add translation.
                self._pop_state()
                self._status = None
                return
            self._strokes = tuple(
                s.rtfcre
                # Ignore add translation strokes.
                for t in state.translations[:-1]
                for s in t.strokes
            )
            self._clear_state(undo=True)
            self._translation = ""
            self._engine.hook_connect("send_string", self.send_string)
            self._engine.hook_connect("send_backspaces", self.send_backspaces)
            self._status = "translations"
        elif self._status == "translations":
            state = self._get_state()[0]
            self._engine.hook_disconnect("send_string", self.send_string)
            self._engine.hook_disconnect("send_backspaces", self.send_backspaces)
            self._engine.add_translation(self._strokes, self._translation.strip())
            self._pop_state()
            self._status = None



from collections import deque
import sys
import re

from rtf_tokenize import RtfTokenizer

from plover import log


class RtfParseError(Exception):
    def __init__(self, lnum, cnum, fmt, *fmt_args):
        msg = "line %u, column %u: %s" % (lnum + 1, cnum + 1, fmt % fmt_args)
        super().__init__(msg)


class BadRtfError(Exception):
    def __init__(self, fmt, *fmt_args):
        msg = fmt % fmt_args
        super().__init__(msg)


def finalize_translation(text):
    if not text:
        return text
    # caseCATalyst doesn't put punctuation in \cxp: treat any isolated
    # punctuation at the beginning of the translation as special.
    if text[0] in ".?!:;," and text[1:] in ("", " "):
        return "{" + text[0] + "}" + text[1:]
    left_ws = len(text) - len(text.lstrip())
    if left_ws > 1:
        text = "{^" + text[:left_ws] + "^}" + text[left_ws:]
    right_ws = len(text) - len(text.rstrip())
    if right_ws > 1:
        text = text[:-right_ws] + "{^" + text[-right_ws:] + "^}"
    return text


def parse_rtfcre(text, normalize=lambda s: s, skip_errors=True):
    not_text = r"\{}"
    style_rx = re.compile("s[0-9]+")
    tokenizer = RtfTokenizer(text)
    next_token = tokenizer.next_token
    rewind_token = tokenizer.rewind_token
    # Check header.
    if next_token() != "{" or next_token() != r"\rtf1":
        raise BadRtfError("invalid header")
    # Parse header/document.
    g_destination, g_text = "rtf1", ""
    group_stack = deque()
    stylesheet = {}
    steno = None
    while True:
        token = next_token()
        # EOF.
        if token is None:
            err = RtfParseError(
                tokenizer.lnum, tokenizer.cnum, "unexpected end of file"
            )
            if not skip_errors:
                raise err
            log.error("%s", err)
            break
        # Group start.
        if token == "{":
            # Always rewind the last token?
            rewind = False
            # Is it an ignored group?
            is_ignored = False
            destination = None
            token = next_token()
            # Ignored?
            if token == r"\*":
                token = next_token()
                is_ignored = True
            # Destination?
            if token[0] == "\\":
                destination = token[1:]
                # Steno.
                if destination == "cxs":
                    if group_stack:
                        err = RtfParseError(
                            tokenizer.lnum,
                            tokenizer.cnum,
                            "starting new mapping, but previous is unfinished",
                        )
                        if not skip_errors:
                            raise err
                        log.error("%s", err)
                        # Simulate missing group end(s).
                        assert group_stack[0][0] == "rtf1"
                        rewind_token(token)
                        if is_ignored:
                            rewind_token(r"\*")
                        rewind_token("{")
                        for __ in range(len(group_stack)):
                            rewind_token("}")
                        continue
                    if steno is not None:
                        yield normalize(steno), finalize_translation(g_text)
                        steno = None
                    is_ignored = False
                    # Reset text.
                    g_text = ""
                elif destination in {
                    # Fingerspelling.
                    "cxfing",
                    # Stenovations extensions...
                    "cxsvatdictflags",
                    # Plover macro.
                    "cxplovermacro",
                    # Plover meta.
                    "cxplovermeta",
                }:
                    is_ignored = False
                elif style_rx.fullmatch(destination):
                    pass
                else:
                    # In the case of e.g. `{\par...`,
                    # `\par` must be handled as a
                    # control word.
                    rewind = True
            else:
                rewind = True
            if is_ignored:
                # Skip ignored content.
                stack_depth = 1
                while True:
                    token = next_token()
                    if token is None:
                        err = RtfParseError(
                            tokenizer.lnum, tokenizer.cnum, "unexpected end of file"
                        )
                        if not skip_errors:
                            raise err
                        log.error("%s", err)
                        break
                    if token == "{":
                        stack_depth += 1
                    elif token == "}":
                        stack_depth -= 1
                        if not stack_depth:
                            break
                if stack_depth:
                    break
                continue
            group_stack.append((g_destination, g_text))
            g_destination, g_text = destination, ""
            if rewind:
                rewind_token(token)
            continue
        # Group end.
        if token == "}":
            if not group_stack:
                token = next_token()
                if token is None:
                    # The end...
                    break
                err = RtfParseError(
                    tokenizer.lnum,
                    tokenizer.cnum,
                    "expected end of file, got: %r",
                    token[0],
                )
                if not skip_errors:
                    raise err
                log.error("%s", err)
                rewind_token(token)
                continue
            # Steno.
            if g_destination == "cxs":
                steno = g_text
                text = ""
            # Punctuation.
            elif g_destination == "cxp":
                text = g_text.strip()
                if text in {".", "!", "?", ",", ";", ":"}:
                    text = "{" + text + "}"
                elif text == "'":
                    text = "{^'}"
                elif text in ("-", "/"):
                    text = "{^" + text + "^}"
                else:
                    # Show unknown punctuation as given.
                    text = "{^" + g_text + "^}"
            # Stenovations extensions...
            elif g_destination == "cxsvatdictflags":
                if "N" in g_text:
                    text = "{-|}"
                else:
                    text = ""
            # Fingerspelling.
            elif g_destination == "cxfing":
                text = "{&" + g_text + "}"
            # Plover macro.
            elif g_destination == "cxplovermacro":
                text = "=" + g_text
            # Plover meta.
            elif g_destination == "cxplovermeta":
                text = "{" + g_text + "}"
            # Style declaration.
            elif (
                g_destination is not None
                and style_rx.fullmatch(g_destination)
                and group_stack[-1][0] == "stylesheet"
            ):
                stylesheet[g_destination] = g_text
            else:
                text = g_text
            g_destination, g_text = group_stack.pop()
            g_text += text
            continue
        # Control char/word.
        if token[0] == "\\":
            ctrl = token[1:]
            text = {
                # Ignore.
                "*": "",
                # Hard space.
                "~": "{^ ^}",
                # Non-breaking hyphen.
                "_": "{^-^}",
                # Escaped newline: \par.
                "": "\n\n",
                "\n": "\n\n",
                "\r": "\n\n",
                # Escaped characters.
                "\\": "\\",
                "{": "{",
                "}": "}",
                "-": "-",
                # Line break.
                "line": "\n",
                # Paragraph break.
                "par": "\n\n",
                # Tab.
                "tab": "\t",
                # Force Cap.
                "cxfc": "{-|}",
                # Force Lower Case.
                "cxfl": "{>}",
            }.get(ctrl)
            if text is not None:
                g_text += text
            # Delete Spaces.
            elif ctrl == "cxds":
                token = next_token()
                if token is None or token[0] in not_text:
                    g_text += "{^}"
                    rewind_token(token)
                else:
                    text = token
                    token = next_token()
                    if token == r"\cxds":
                        # Infix
                        g_text += "{^" + text + "^}"
                    else:
                        # Prefix.
                        g_text += "{^" + text + "}"
                        rewind_token(token)
            # Delete Last Stroke.
            elif ctrl == "cxdstroke":
                g_text = "=undo"
            # Fingerspelling.
            elif ctrl == "cxfing":
                token = next_token()
                if token is None or token[0] in not_text:
                    err = RtfParseError(
                        tokenizer.lnum, tokenizer.cnum, "expected text, got: %r", token
                    )
                    if not skip_errors:
                        raise err
                    log.error("%s", err)
                    rewind_token(token)
                else:
                    g_text += "{&" + token + "}"
            elif style_rx.fullmatch(ctrl):
                # Workaround for caseCATalyst declaring
                # new styles without a preceding \par.
                if not g_text.endswith("\n\n"):
                    g_text += "\n\n"
                # Indent continuation styles.
                if stylesheet.get(ctrl, "").startswith("Contin"):
                    g_text += "    "
            continue
        # Text.
        text = token
        token = next_token()
        if token == r"\cxds":
            # Suffix.
            text = "{" + text + "^}"
        else:
            rewind_token(token)
        g_text += text
    if steno is not None:
        yield normalize(steno), finalize_translation(g_text)


def main(todo, filename):
    with open(filename, "rb") as fp:
        text = fp.read().decode("cp1252")
    if todo == "tokenize":
        next_token = RtfTokenizer(text).next_token
        while next_token() is not None:
            pass
    elif todo == "parse":
        for __ in parse_rtfcre(text):
            pass
    elif todo == "dump_tokenize":
        tokenizer = RtfTokenizer(text)
        while True:
            token = tokenizer.next_token()
            if token is None:
                break
            print("%3u:%-3u %r" % (tokenizer.lnum + 1, tokenizer.cnum + 1, token))
    elif todo == "dump_parse":
        for mapping in parse_rtfcre(text):
            print(mapping)
    else:
        raise ValueError(todo)


if __name__ == "__main__":
    assert len(sys.argv) == 3
    main(sys.argv[1], sys.argv[2])



from plover.translation import Translation
from plover.steno import Stroke
from plover import system


def toggle_asterisk(translator, stroke, cmdline):
    assert not cmdline
    # Toggle asterisk of previous stroke
    translations = translator.get_state().translations
    if not translations:
        return
    t = translations[-1]
    translator.untranslate_translation(t)
    keys = set(t.strokes[-1].steno_keys)
    if "*" in keys:
        keys.remove("*")
    else:
        keys.add("*")
    translator.translate_stroke(Stroke(keys))


def delete_space(translator, stroke, cmdline):
    assert not cmdline
    # Retrospective delete space
    translations = translator.get_state().translations
    if len(translations) < 2:
        return
    replaced = translations[-2:]
    if replaced[1].is_retrospective_command:
        return
    english = []
    for t in replaced:
        if t.english is not None:
            english.append(t.english)
        elif len(t.rtfcre) == 1 and t.rtfcre[0].isdigit():
            english.append("{&%s}" % t.rtfcre[0])
    if len(english) > 1:
        t = Translation([stroke], "{^~|^}".join(english))
        t.replaced = replaced
        t.is_retrospective_command = True
        translator.translate_translation(t)


def insert_space(translator, stroke, cmdline):
    assert not cmdline
    # Retrospective insert space
    translations = translator.get_state().translations
    if not translations:
        return
    replaced = translations[-1]
    if replaced.is_retrospective_command:
        return
    lookup_stroke = replaced.strokes[-1]
    english = [t.english or "/".join(t.rtfcre) for t in replaced.replaced]
    if english:
        english.append(
            translator.lookup([lookup_stroke], system.SUFFIX_KEYS)
            or lookup_stroke.rtfcre
        )
        t = Translation([stroke], " ".join(english))
        t.replaced = [replaced]
        t.is_retrospective_command = True
        translator.translate_translation(t)



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""Unit tests for steno_dictionary.py."""

import pytest

from plover.steno_dictionary import StenoDictionary, StenoDictionaryCollection

from plover_build_utils.testing import dictionary_test


def test_dictionary_collection():
    d1 = StenoDictionary()
    d1[("S",)] = "a"
    d1[("T",)] = "b"
    d1[("H",)] = "{PLOVER:deleted}"
    d1[("R",)] = "e"
    d1.path = "d1"
    d2 = StenoDictionary()
    d2[("S",)] = "c"
    d2[("W",)] = "d"
    d2[("H",)] = "f"
    d2[("R",)] = "{plover:deleted}"
    d2.path = "d2"
    dc = StenoDictionaryCollection([d2, d1])
    assert dc.lookup(("S",)) == "c"
    assert dc.lookup(("W",)) == "d"
    assert dc.lookup(("T",)) == "b"
    assert dc.lookup(("H",)) == "f"
    assert dc.lookup(("R",)) == None
    assert dc.lookup_from_all(("S",)) == [("c", d2), ("a", d1)]
    assert dc.lookup_from_all(("W",)) == [("d", d2)]
    assert dc.lookup_from_all(("T",)) == [("b", d1)]
    assert dc.lookup_from_all(("H",)) == [("f", d2), ("{PLOVER:deleted}", d1)]
    assert dc.lookup_from_all(("R",)) == [("{plover:deleted}", d2), ("e", d1)]
    f = lambda k, v: v == "c"
    dc.add_filter(f)
    assert dc.lookup(("S",)) == "a"
    assert dc.raw_lookup(("S",)) == "c"
    assert dc.lookup_from_all(("S",)) == [("a", d1)]
    g = lambda k, v: v == "a"
    dc.add_filter(g)
    assert dc.lookup(("S",)) is None
    assert dc.lookup_from_all(("S",)) == []
    assert dc.raw_lookup(("S",)) == "c"
    assert dc.lookup(("W",)) == "d"
    assert dc.lookup(("T",)) == "b"
    assert dc.raw_lookup(("W",)) == "d"
    assert dc.raw_lookup(("T",)) == "b"
    assert dc.raw_lookup_from_all(("S",)) == [("c", d2), ("a", d1)]
    assert dc.raw_lookup_from_all(("W",)) == [("d", d2)]
    assert dc.raw_lookup_from_all(("T",)) == [("b", d1)]
    assert dc.reverse_lookup("c") == {("S",)}

    dc.remove_filter(f)
    dc.remove_filter(g)
    assert dc.lookup(("S",)) == "c"
    assert dc.lookup(("W",)) == "d"
    assert dc.lookup(("T",)) == "b"
    assert dc.lookup(("H",)) == "f"
    assert dc.lookup(("R",)) == None
    assert dc.lookup_from_all(("S",)) == [("c", d2), ("a", d1)]
    assert dc.lookup_from_all(("W",)) == [("d", d2)]
    assert dc.lookup_from_all(("T",)) == [("b", d1)]
    assert dc.lookup_from_all(("H",)) == [("f", d2), ("{PLOVER:deleted}", d1)]
    assert dc.lookup_from_all(("R",)) == [("{plover:deleted}", d2), ("e", d1)]

    assert dc.reverse_lookup("c") == {("S",)}

    dc.set(("S",), "e")
    assert dc.lookup(("S",)) == "e"
    assert d2[("S",)] == "e"

    dc.set(("S",), "f", path="d1")
    assert dc.lookup(("S",)) == "e"
    assert d1[("S",)] == "f"
    assert d2[("S",)] == "e"

    # Iterating on a StenoDictionaryCollection is
    # the same as iterating on its dictionaries' paths.
    assert list(dc) == ["d2", "d1"]

    # Test get and [].
    assert dc.get("d1") == d1
    assert dc["d1"] == d1
    assert dc.get("invalid") is None
    with pytest.raises(KeyError):
        dc["invalid"]


def test_dictionary_collection_writeable():
    d1 = StenoDictionary()
    d1[("S",)] = "a"
    d1[("T",)] = "b"
    d2 = StenoDictionary()
    d2[("S",)] = "c"
    d2[("W",)] = "d"
    d2.readonly = True
    dc = StenoDictionaryCollection([d2, d1])
    assert dc.first_writable() == d1
    dc.set(("S",), "A")
    assert d1[("S",)] == "A"
    assert d2[("S",)] == "c"


def test_dictionary_collection_longest_key():
    k1 = ("S",)
    k2 = ("S", "T")
    k3 = ("S", "T", "R")

    dc = StenoDictionaryCollection()
    assert dc.longest_key == 0

    d1 = StenoDictionary()
    d1.path = "d1"
    d1[k1] = "a"

    dc.set_dicts([d1])
    assert dc.longest_key == 1

    d1[k2] = "a"
    assert dc.longest_key == 2

    d2 = StenoDictionary()
    d2.path = "d2"
    d2[k3] = "c"

    dc.set_dicts([d2, d1])
    assert dc.longest_key == 3

    del d1[k2]
    assert dc.longest_key == 3

    d2.enabled = False
    assert dc.longest_key == 1

    d1.enabled = False
    assert dc.longest_key == 0

    d2.enabled = True
    assert dc.longest_key == 3

    d1.enabled = True
    assert dc.longest_key == 3

    dc.set_dicts([d1])
    assert dc.longest_key == 1

    dc.set_dicts([])
    assert dc.longest_key == 0


def test_casereverse_lookup():
    dc = StenoDictionaryCollection()

    d1 = StenoDictionary()
    d1[("PWAOUFL",)] = "beautiful"
    d1[("WAOUFL",)] = "beAuTIFul"

    d2 = StenoDictionary()
    d2[("PW-FL",)] = "BEAUTIFUL"

    d3 = StenoDictionary()
    d3[("WAOUFL",)] = "not beautiful"

    dc.set_dicts([d1, d2, d3])

    assert dc.casereverse_lookup("beautiful") == {"beautiful", "BEAUTIFUL", "beAuTIFul"}


def test_reverse_lookup():
    dc = StenoDictionaryCollection()

    d1 = StenoDictionary()
    d1[("PWAOUFL",)] = "beautiful"
    d1[("WAOUFL",)] = "beautiful"

    d2 = StenoDictionary()
    d2[("PW-FL",)] = "beautiful"

    d3 = StenoDictionary()
    d3[("WAOUFL",)] = "not beautiful"

    # Simple test.
    dc.set_dicts([d1])
    assert dc.reverse_lookup("beautiful") == {("PWAOUFL",), ("WAOUFL",)}

    # No duplicates.
    d2_copy = StenoDictionary()
    d2_copy.update(d2)
    dc.set_dicts([d2_copy, d2])
    assert dc.reverse_lookup("beautiful") == {("PW-FL",)}

    # Don't stop at the first dictionary with matches.
    dc.set_dicts([d2, d1])
    assert dc.reverse_lookup("beautiful") == {("PW-FL",), ("PWAOUFL",), ("WAOUFL",)}

    # Ignore keys overridden by a higher precedence dictionary.
    dc.set_dicts([d3, d2, d1])
    assert dc.reverse_lookup("beautiful") == {("PW-FL",), ("PWAOUFL",)}


def test_dictionary_enabled():
    dc = StenoDictionaryCollection()
    d1 = StenoDictionary()
    d1.path = "d1"
    d1[("TEFT",)] = "test1"
    d1[("TEFGT",)] = "Testing"
    d2 = StenoDictionary()
    d2[("TEFT",)] = "test2"
    d2[("TEFT", "-G")] = "Testing"
    d2.path = "d2"
    dc.set_dicts([d2, d1])
    assert dc.lookup(("TEFT",)) == "test2"
    assert dc.raw_lookup(("TEFT",)) == "test2"
    assert dc.casereverse_lookup("testing") == {"Testing"}
    assert dc.reverse_lookup("Testing") == {("TEFT", "-G"), ("TEFGT",)}
    d2.enabled = False
    assert dc.lookup(("TEFT",)) == "test1"
    assert dc.raw_lookup(("TEFT",)) == "test1"
    assert dc.casereverse_lookup("testing") == {"Testing"}
    assert dc.reverse_lookup("Testing") == {("TEFGT",)}
    d1.enabled = False
    assert dc.lookup(("TEST",)) is None
    assert dc.raw_lookup(("TEFT",)) is None
    assert dc.casereverse_lookup("testing") == set()
    assert dc.reverse_lookup("Testing") == set()


@dictionary_test
class TestStenoDictionary:
    class DICT_CLASS(StenoDictionary):
        def _load(self, filename):
            pass

    DICT_EXTENSION = "dict"
    DICT_SAMPLE = b""


@dictionary_test
class TestReadOnlyStenoDictionary:
    class DICT_CLASS(StenoDictionary):
        readonly = True

        def _load(self, filename):
            pass

    DICT_EXTENSION = "dict"
    DICT_SAMPLE = b""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author: @abarnert, @willwade, and @morinted
# Code taken and modified from
# <https://github.com/willwade/PyUserInput/blob/master/pykeyboard/mac_keycode.py>
# <https://stackoverflow.com/questions/1918841/how-to-convert-ascii-character-to-cgkeycode>

from threading import Thread
import ctypes
import ctypes.util
import re
import struct
import unicodedata

from PyObjCTools import AppHelper
import AppKit
import Foundation

from plover import log
from plover.key_combo import CHAR_TO_KEYNAME
from plover.misc import popcount_8


carbon_path = ctypes.util.find_library("Carbon")
carbon = ctypes.cdll.LoadLibrary(carbon_path)

CFIndex = ctypes.c_int64


class CFRange(ctypes.Structure):
    _fields_ = [("loc", CFIndex), ("len", CFIndex)]


carbon.TISCopyCurrentKeyboardInputSource.argtypes = []
carbon.TISCopyCurrentKeyboardInputSource.restype = ctypes.c_void_p
carbon.TISCopyCurrentASCIICapableKeyboardLayoutInputSource.argtypes = []
carbon.TISCopyCurrentASCIICapableKeyboardLayoutInputSource.restype = ctypes.c_void_p

carbon.TISGetInputSourceProperty.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
carbon.TISGetInputSourceProperty.restype = ctypes.c_void_p
carbon.LMGetKbdType.argtypes = []
carbon.LMGetKbdType.restype = ctypes.c_uint32
carbon.CFDataGetLength.argtypes = [ctypes.c_void_p]
carbon.CFDataGetLength.restype = ctypes.c_uint64
carbon.CFDataGetBytes.argtypes = [ctypes.c_void_p, CFRange, ctypes.c_void_p]
carbon.CFDataGetBytes.restype = None
carbon.CFRelease.argtypes = [ctypes.c_void_p]
carbon.CFRelease.restype = None

kTISPropertyUnicodeKeyLayoutData = ctypes.c_void_p.in_dll(
    carbon, "kTISPropertyUnicodeKeyLayoutData"
)
kTISPropertyInputSourceID = ctypes.c_void_p.in_dll(carbon, "kTISPropertyInputSourceID")
kTISPropertyInputSourceIsASCIICapable = ctypes.c_void_p.in_dll(
    carbon, "kTISPropertyInputSourceIsASCIICapable"
)
COMMAND = "⌘"
SHIFT = "⇧"
OPTION = "⌥"
CONTROL = "⌃"
CAPS = "⇪"
UNKNOWN = "?"
UNKNOWN_L = "?_L"

KEY_CODE_VISUALIZATION = """
┌───────────────────Layout of Apple Extended Keyboard II───────────────────────┐
│ 53  122 120 99 118  96 97 98 100  101 109 103 111   105 107 113              │
│                                                                              │
│ 50─── 18 19 20 21 23 22 26 28 25 29 27 24 ─────51   114 115 116  71 81 75 67 │
│ 48──── 12 13 14 15 17 16 32 34 31 35 33 30 ────42   117 119 121  89 91 92 78 │
│ 57───── 0  1  2  3  5  4  38 40 37 41 39 ──────36                86 87 88 69 │
│ 56────── 6  7  8  9  11 45 46  43 47 44 ───────56       126      83 84 85 76 │
│ 59── 58─ 55─ ──────────49───────── ─55 ──58 ───59   123 125 124  82─── 65 76 │
└──────────────────────────────────────────────────────────────────────────────┘
Valid keycodes not visible on layout:
  10, 52, 54, 60-64, 66, 68, 70, 72-74, 77, 79, 80,
  90, 93-95, 102, 104, 106, 108, 110, 112, 127
"""

SPECIAL_KEY_NAMES = {
    "\x1b": "Esc",  # Will map both Esc and Clear (key codes 53 and 71)
    "\xa0": "nbsp",
    "\x08": "Bksp",
    "\x05": "Help",
    "\x01": "Home",
    "\x7f": "Del",
    "\x04": "End",
    "\x0c": "PgDn",
    "\x0b": "PgUp",  # \x0b is also the clear character signal
}

DEFAULT_SEQUENCE = ((None, 0),)


def is_printable(string):
    for character in string:
        category = unicodedata.category(character)
        if category[0] in "C":
            # Exception: the "Apple" character that most Mac layouts have
            return False if string != "" else True
        elif category == "Zs" and character != " ":
            return False
        elif category in "Zl, Zp":
            return False
    return True


def get_printable_string(s):
    if s is None:
        return "None"
    return (
        s
        if is_printable(s)
        else SPECIAL_KEY_NAMES.setdefault(s, s.encode("unicode_escape").decode("utf-8"))
    )


class KeyboardLayout:
    def __init__(self, watch_layout=True):
        self._char_to_key_sequence = None
        self._key_sequence_to_char = None
        self._modifier_masks = None
        self._deadkey_symbol_to_key_sequence = None

        # Spawn a thread that responds to system keyboard layout changes.
        if watch_layout:
            self._watcher = Thread(target=self._layout_watcher, name="LayoutWatcher")
            self._watcher.start()

        self._update_layout()

    def _layout_watcher(self):
        layout = self

        class LayoutWatchingCallback(AppKit.NSObject):
            def layoutChanged_(self, event):
                log.info("Mac keyboard layout changed, updating")
                try:
                    layout._update_layout()
                except:
                    log.warning("error during layout update, ignoring")

        center = Foundation.NSDistributedNotificationCenter.defaultCenter()
        watcher_callback = LayoutWatchingCallback.new()
        center.addObserver_selector_name_object_suspensionBehavior_(
            watcher_callback,
            "layoutChanged:",
            "com.apple.Carbon.TISNotifySelectedKeyboardInputSourceChanged",
            None,
            Foundation.NSNotificationSuspensionBehaviorDeliverImmediately,
        )
        AppHelper.runConsoleEventLoop(installInterrupt=True)

    def _update_layout(self):
        char_to_key_sequence, key_sequence_to_char, modifier_masks = (
            KeyboardLayout._get_layout()
        )
        self._char_to_key_sequence = char_to_key_sequence
        self._key_sequence_to_char = key_sequence_to_char
        self._modifier_masks = modifier_masks
        self._deadkey_symbol_to_key_sequence = self._deadkeys_by_symbols()

    def deadkey_symbol_to_key_sequence(self, symbol):
        return self._deadkey_symbol_to_key_sequence.get(symbol, DEFAULT_SEQUENCE)

    def key_code_to_char(self, key_code, modifier=0):
        """Provide a key code and a modifier and it provides the character"""
        return get_printable_string(self._key_sequence_to_char[key_code, modifier])

    def char_to_key_sequence(self, char):
        """Finds the key code and modifier sequence for the character"""
        key_sequence = self._char_to_key_sequence.get(char, DEFAULT_SEQUENCE)
        if not isinstance(key_sequence[0], tuple):
            key_sequence = (key_sequence,)
        return key_sequence

    def _deadkeys_by_symbols(self):
        # We store deadkeys as "characters"; dkX, where X is the symbol.
        symbols = {
            "`": "`",
            "´": "´",
            "^": ("^", "ˆ"),
            "~": ("~", "˜"),
            "¨": "¨",
        }
        deadkeys_by_symbol = {}
        for symbol, equivalent_symbols in symbols.items():
            for equivalent_symbol in equivalent_symbols:
                sequence = self.char_to_key_sequence("dk%s" % equivalent_symbol)
                if sequence[0][0] is not None:
                    deadkeys_by_symbol[symbol] = sequence
        return deadkeys_by_symbol

    def format_modifier_header(self):
        modifiers = (
            "| {}\t".format(KeyboardLayout._modifier_string(mod)).expandtabs(8)
            for mod in sorted(self._modifier_masks.values())
        )
        header = "Keycode\t{}".format("".join(modifiers))
        return "%s\n%s" % (header, re.sub(r"[^|]", "-", header))

    def format_keycode_keys(self, keycode):
        """Returns all the variations of the keycode with modifiers"""
        keys = (
            "| {}\t".format(
                get_printable_string(self._key_sequence_to_char[keycode, mod])
            ).expandtabs(8)
            for mod in sorted(self._modifier_masks.values())
        )

        return "{}\t{}".format(keycode, "".join(keys)).expandtabs(8)

    @staticmethod
    def _modifier_dictionary(modifier_mask):
        """Provide a dictionary of active modifier keys from mod mask"""
        modifiers = {
            SHIFT: False,
            COMMAND: False,
            CONTROL: False,
            OPTION: False,
            CAPS: False,
            UNKNOWN: False,
            UNKNOWN_L: False,
        }
        if modifier_mask & 16:
            modifiers[CONTROL] = True
        if modifier_mask & 8:
            modifiers[OPTION] = True
        if modifier_mask & 4:
            modifiers[CAPS] = True
        if modifier_mask & 2:
            modifiers[SHIFT] = True
        if modifier_mask & 1:
            modifiers[COMMAND] = True
        return modifiers

    @staticmethod
    def _modifier_string(modifier_mask):
        """Turn modifier mask into string representing modifiers"""
        s = ""
        modifiers = KeyboardLayout._modifier_dictionary(modifier_mask)
        for key in modifiers:
            s += key if modifiers[key] else ""
        return s

    @staticmethod
    def _get_layout():
        keyboard_input_source = carbon.TISCopyCurrentKeyboardInputSource()

        layout_source = carbon.TISGetInputSourceProperty(
            keyboard_input_source, kTISPropertyUnicodeKeyLayoutData
        )
        # Some keyboard layouts don't return UnicodeKeyLayoutData so we turn to a different source.
        if layout_source is None:
            carbon.CFRelease(keyboard_input_source)
            keyboard_input_source = (
                carbon.TISCopyCurrentASCIICapableKeyboardLayoutInputSource()
            )
            layout_source = carbon.TISGetInputSourceProperty(
                keyboard_input_source, kTISPropertyUnicodeKeyLayoutData
            )
        layout_size = carbon.CFDataGetLength(layout_source)
        layout_buffer = ctypes.create_string_buffer(b"\000" * layout_size)
        carbon.CFDataGetBytes(
            layout_source, CFRange(0, layout_size), ctypes.byref(layout_buffer)
        )
        keyboard_type = carbon.LMGetKbdType()
        parsed_layout = KeyboardLayout._parse_layout(layout_buffer, keyboard_type)
        carbon.CFRelease(keyboard_input_source)
        return parsed_layout

    @staticmethod
    def _parse_layout(buf, ktype):
        hf, dv, featureinfo, ktcount = struct.unpack_from("HHII", buf)
        offset = struct.calcsize("HHII")
        ktsize = struct.calcsize("IIIIIII")
        kts = [
            struct.unpack_from("IIIIIII", buf, offset + ktsize * i)
            for i in range(ktcount)
        ]
        for i, kt in enumerate(kts):
            if kt[0] <= ktype <= kt[1]:
                kentry = i
                break
        else:
            kentry = 0
        ktf, ktl, modoff, charoff, sroff, stoff, seqoff = kts[kentry]

        # Modifiers
        mf, deftable, mcount = struct.unpack_from("HHI", buf, modoff)
        modtableoff = modoff + struct.calcsize("HHI")
        modtables = struct.unpack_from("B" * mcount, buf, modtableoff)
        modifier_masks = {}
        for i, table in enumerate(modtables):
            modifier_masks.setdefault(table, i)

        # Sequences
        sequences = []
        if seqoff:
            sf, scount = struct.unpack_from("HH", buf, seqoff)
            seqtableoff = seqoff + struct.calcsize("HH")
            lastoff = -1
            for soff in struct.unpack_from("H" * scount, buf, seqtableoff):
                if lastoff >= 0:
                    sequences.append(
                        buf[seqoff + lastoff : seqoff + soff].decode("utf-16")
                    )
                lastoff = soff

        def lookupseq(key):
            if key >= 0xFFFE:
                return None
            if key & 0xC000:
                seq = key & ~0xC000
                if seq < len(sequences):
                    return sequences[seq]
            return chr(key)

        # Dead keys
        deadkeys = []
        if sroff:
            srf, srcount = struct.unpack_from("HH", buf, sroff)
            srtableoff = sroff + struct.calcsize("HH")
            for recoff in struct.unpack_from("I" * srcount, buf, srtableoff):
                cdata, nextstate, ecount, eformat = struct.unpack_from(
                    "HHHH", buf, recoff
                )
                recdataoff = recoff + struct.calcsize("HHHH")
                edata = buf[recdataoff : recdataoff + 4 * ecount]
                deadkeys.append((cdata, nextstate, ecount, eformat, edata))

        if stoff:
            stf, stcount = struct.unpack_from("HH", buf, stoff)
            sttableoff = stoff + struct.calcsize("HH")
            dkterms = struct.unpack_from("H" * stcount, buf, sttableoff)
        else:
            dkterms = []

        # Get char tables
        char_to_key_sequence = {}
        key_sequence_to_char = {}

        deadkey_state_to_key_sequence = {}
        key_sequence_to_deadkey_state = {}

        def favored_modifiers(modifier_a, modifier_b):
            """0 if they are equal, 1 if a is better, -1 if b is better"""
            a_favored_over_b = 0
            modifiers_a = KeyboardLayout._modifier_dictionary(modifier_a)
            modifiers_b = KeyboardLayout._modifier_dictionary(modifier_b)
            count_a = popcount_8(modifier_a)
            count_b = popcount_8(modifier_b)

            # Favor no CAPS modifier
            if modifiers_a[CAPS] and not modifiers_b[CAPS]:
                a_favored_over_b = -1
            elif modifiers_b[CAPS] and not modifiers_a[CAPS]:
                a_favored_over_b = 1
            # Favor no command modifier
            elif modifiers_a[COMMAND] and not modifiers_b[COMMAND]:
                a_favored_over_b = -1
            elif modifiers_b[COMMAND] and not modifiers_a[COMMAND]:
                a_favored_over_b = 1
            # Favor no control modifier
            elif modifiers_a[CONTROL] and not modifiers_b[CONTROL]:
                a_favored_over_b = -1
            elif modifiers_b[CONTROL] and not modifiers_a[CONTROL]:
                a_favored_over_b = 1
            # Finally, favor fewer modifiers
            elif count_a > count_b:
                a_favored_over_b = -1
            elif count_b > count_a:
                a_favored_over_b = 1
            return a_favored_over_b

        def save_shortest_key_sequence(character, new_sequence):
            current_sequence = char_to_key_sequence.setdefault(character, new_sequence)
            if current_sequence != new_sequence:
                # Convert responses to tuples...
                if not isinstance(current_sequence[0], tuple):
                    current_sequence = (current_sequence,)
                if not isinstance(new_sequence[0], tuple):
                    new_sequence = (new_sequence,)

                first_current_better = favored_modifiers(
                    current_sequence[-1][1], new_sequence[-1][1]
                )
                last_current_better = favored_modifiers(
                    current_sequence[0][1], new_sequence[0][1]
                )

                # Favor key sequence with best modifiers (avoids a short sequence with awful modifiers)
                # e.g. ABC Extended wants ¯ from (⌘⌥a,) instead of the saner (⌥a, space)
                if first_current_better == last_current_better != 0:
                    if first_current_better < 0:
                        char_to_key_sequence[character] = new_sequence
                # Favor shortest sequence (fewer separate key presses)
                elif len(current_sequence) < len(new_sequence):
                    pass
                elif len(new_sequence) < len(current_sequence):
                    char_to_key_sequence[character] = new_sequence[0]
                # Favor fewer modifiers on last item
                elif last_current_better < 0:
                    char_to_key_sequence[character] = new_sequence
                # Favor lower modifiers on first item if last item is the same
                elif last_current_better == 0 and first_current_better < 0:
                    char_to_key_sequence[character] = new_sequence

        def lookup_and_add(key, j, mod):
            ch = lookupseq(key)
            save_shortest_key_sequence(ch, (j, mod))
            key_sequence_to_char[j, mod] = ch

        cf, csize, ccount = struct.unpack_from("HHI", buf, charoff)
        chartableoff = charoff + struct.calcsize("HHI")

        for i, table_offset in enumerate(
            struct.unpack_from("I" * ccount, buf, chartableoff)
        ):
            mod = modifier_masks[i]
            for j, key in enumerate(struct.unpack_from("H" * csize, buf, table_offset)):
                if key == 65535:
                    key_sequence_to_char[j, mod] = "mod"
                elif key >= 0xFFFE:
                    key_sequence_to_char[j, mod] = "<{}>".format(key)
                elif key & 0x0C000 == 0x4000:
                    dead = key & ~0xC000
                    if dead < len(deadkeys):
                        cdata, nextstate, ecount, eformat, edata = deadkeys[dead]
                        if eformat == 0 and nextstate:
                            # Initial: symbols, e.g. `, ~, ^
                            new_deadkey = (j, mod)
                            current_deadkey = deadkey_state_to_key_sequence.setdefault(
                                nextstate, new_deadkey
                            )
                            if new_deadkey != current_deadkey:
                                if (
                                    favored_modifiers(
                                        current_deadkey[1], new_deadkey[1]
                                    )
                                    < 0
                                ):
                                    deadkey_state_to_key_sequence[nextstate] = (
                                        new_deadkey
                                    )
                            if nextstate - 1 < len(dkterms):
                                base_key = lookupseq(dkterms[nextstate - 1])
                                dead_key_name = "dk{}".format(base_key)
                            else:
                                dead_key_name = "dk#{}".format(nextstate)
                            key_sequence_to_char[j, mod] = dead_key_name
                            save_shortest_key_sequence(dead_key_name, (j, mod))
                        elif eformat == 1 or (eformat == 0 and not nextstate):
                            # Terminal: letters, e.g. a, e, o, A, E, O
                            key_sequence_to_deadkey_state[j, mod] = deadkeys[dead]

                            lookup_and_add(cdata, j, mod)
                        elif eformat == 2:  # range
                            # TODO!
                            pass
                else:
                    lookup_and_add(key, j, mod)

        for key, dead in key_sequence_to_deadkey_state.items():
            j, mod = key
            cdata, nextstate, ecount, eformat, edata = dead
            entries = [struct.unpack_from("HH", edata, i * 4) for i in range(ecount)]
            for state, key in entries:
                # Ignore if unknown state...
                if state in deadkey_state_to_key_sequence:
                    dj, dmod = deadkey_state_to_key_sequence[state]
                    ch = lookupseq(key)
                    save_shortest_key_sequence(ch, ((dj, dmod), (j, mod)))
                    key_sequence_to_char[(dj, dmod), (j, mod)] = ch

        char_to_key_sequence["\n"] = (36, 0)
        char_to_key_sequence["\r"] = (36, 0)
        char_to_key_sequence["\t"] = (48, 0)

        return char_to_key_sequence, key_sequence_to_char, modifier_masks


if __name__ == "__main__":
    layout = KeyboardLayout(False)
    print(KEY_CODE_VISUALIZATION)
    print()
    print(layout.format_modifier_header())
    for keycode in range(127):
        print(layout.format_keycode_keys(keycode))
    print()
    unmapped_characters = []
    for char, keyname in sorted(CHAR_TO_KEYNAME.items()):
        sequence = []
        for code, mod in layout.char_to_key_sequence(char):
            if code is not None:
                sequence.append(
                    (
                        code,
                        "{}{}".format(
                            layout._modifier_string(mod),
                            layout.key_code_to_char(code, 0),
                        ),
                    )
                )
            else:
                unmapped_characters.append(char)
        if sequence:
            print(
                "Name:\t\t{}\nCharacter:\t{}\nSequence:\t‘{}’\nBase codes:\t‘{}’\n".format(
                    keyname,
                    char,
                    "’, ‘".join(combo[1] for combo in sequence),
                    "’, ‘".join(str(combo[0]) for combo in sequence),
                )
            )
    print(
        "No mapping on this layout for characters: ‘{}’".format(
            "’, ‘".join(unmapped_characters)
        )
    )



# -*- coding: utf-8 -*-

from collections import defaultdict, namedtuple
from ctypes import windll, wintypes
import codecs
import ctypes
import sys

from plover.key_combo import CHAR_TO_KEYNAME, add_modifiers_aliases
from plover.misc import popcount_8

from .wmctrl import GetForegroundWindow


GetKeyboardLayout = windll.user32.GetKeyboardLayout
GetKeyboardLayout.argtypes = [
    wintypes.DWORD,  # idThread
]
GetKeyboardLayout.restype = wintypes.HKL

GetWindowThreadProcessId = windll.user32.GetWindowThreadProcessId
GetWindowThreadProcessId.argtypes = [
    wintypes.HWND,  # hWnd
    wintypes.LPDWORD,  # lpdwProcessId
]
GetWindowThreadProcessId.restype = wintypes.DWORD

MapVirtualKeyEx = windll.user32.MapVirtualKeyExW
MapVirtualKeyEx.argtypes = [
    wintypes.UINT,  # uCode,
    wintypes.UINT,  # uMapType,
    wintypes.HKL,  # dwhkl
]
MapVirtualKeyEx.restype = wintypes.UINT

ToUnicodeEx = windll.user32.ToUnicodeEx
ToUnicodeEx.argtypes = [
    wintypes.UINT,  # wVirtKey,
    wintypes.UINT,  # wScanCode,
    wintypes.LPBYTE,  # lpKeyState,
    wintypes.LPWSTR,  # pwszBuff,
    ctypes.c_int,  # cchBuff,
    wintypes.UINT,  # wFlags,
    wintypes.HKL,  # dwhkl
]
ToUnicodeEx.restype = ctypes.c_int


def enum(name, items):
    """Helper to simulate an Enum."""
    keys, values = zip(*items)
    t = namedtuple(name, keys)
    return t._make(values)


# Shift state enum. {{{

SHIFT_STATE = enum(
    "SHIFT_STATE",
    (
        (mod, n)
        for n, mod in enumerate(
            "BASE SHIFT CTRL SHIFT_CTRL MENU SHIFT_MENU MENU_CTRL SHIFT_MENU_CTRL".split()
        )
    ),
)


def shift_state_str(ss):
    s = ""
    for mod_state, mod_str in (
        (SHIFT_STATE.SHIFT, "SHIFT"),
        (SHIFT_STATE.CTRL, "CTRL"),
        (SHIFT_STATE.MENU, "MENU"),
    ):
        if (ss & mod_state) != 0:
            s += mod_str + "+"
    return s


# }}}

# Virtual keys enum. {{{
# fmt: off
vk_dict = {
    "LBUTTON"            : 0x01,
    "RBUTTON"            : 0x02,
    "CANCEL"             : 0x03,
    "MBUTTON"            : 0x04,
    "XBUTTON1"           : 0x05,
    "XBUTTON2"           : 0x06,
    "BACK"               : 0x08,
    "TAB"                : 0x09,
    "CLEAR"              : 0x0C,
    "RETURN"             : 0x0D,
    "SHIFT"              : 0x10,
    "CONTROL"            : 0x11,
    "MENU"               : 0x12,
    "PAUSE"              : 0x13,
    "CAPITAL"            : 0x14,
    "KANA"               : 0x15,
    "JUNJA"              : 0x17,
    "FINAL"              : 0x18,
    "HANJA"              : 0x19,
    "ESCAPE"             : 0x1B,
    "CONVERT"            : 0x1C,
    "NONCONVERT"         : 0x1D,
    "ACCEPT"             : 0x1E,
    "MODECHANGE"         : 0x1F,
    "SPACE"              : 0x20,
    "PRIOR"              : 0x21,
    "NEXT"               : 0x22,
    "END"                : 0x23,
    "HOME"               : 0x24,
    "LEFT"               : 0x25,
    "UP"                 : 0x26,
    "RIGHT"              : 0x27,
    "DOWN"               : 0x28,
    "SELECT"             : 0x29,
    "PRINT"              : 0x2A,
    "EXECUTE"            : 0x2B,
    "SNAPSHOT"           : 0x2C,
    "INSERT"             : 0x2D,
    "DELETE"             : 0x2E,
    "HELP"               : 0x2F,
    "LWIN"               : 0x5B,
    "RWIN"               : 0x5C,
    "APPS"               : 0x5D,
    "SLEEP"              : 0x5F,
    "NUMPAD0"            : 0x60,
    "NUMPAD1"            : 0x61,
    "NUMPAD2"            : 0x62,
    "NUMPAD3"            : 0x63,
    "NUMPAD4"            : 0x64,
    "NUMPAD5"            : 0x65,
    "NUMPAD6"            : 0x66,
    "NUMPAD7"            : 0x67,
    "NUMPAD8"            : 0x68,
    "NUMPAD9"            : 0x69,
    "MULTIPLY"           : 0x6A,
    "ADD"                : 0x6B,
    "SEPARATOR"          : 0x6C,
    "SUBTRACT"           : 0x6D,
    "DECIMAL"            : 0x6E,
    "DIVIDE"             : 0x6F,
    "F1"                 : 0x70,
    "F2"                 : 0x71,
    "F3"                 : 0x72,
    "F4"                 : 0x73,
    "F5"                 : 0x74,
    "F6"                 : 0x75,
    "F7"                 : 0x76,
    "F8"                 : 0x77,
    "F9"                 : 0x78,
    "F10"                : 0x79,
    "F11"                : 0x7A,
    "F12"                : 0x7B,
    "F13"                : 0x7C,
    "F14"                : 0x7D,
    "F15"                : 0x7E,
    "F16"                : 0x7F,
    "F17"                : 0x80,
    "F18"                : 0x81,
    "F19"                : 0x82,
    "F20"                : 0x83,
    "F21"                : 0x84,
    "F22"                : 0x85,
    "F23"                : 0x86,
    "F24"                : 0x87,
    "NUMLOCK"            : 0x90,
    "SCROLL"             : 0x91,
    "OEM_NEC_EQUAL"      : 0x92,
    "OEM_FJ_JISHO"       : 0x92,
    "OEM_FJ_MASSHOU"     : 0x93,
    "OEM_FJ_TOUROKU"     : 0x94,
    "OEM_FJ_LOYA"        : 0x95,
    "OEM_FJ_ROYA"        : 0x96,
    "LSHIFT"             : 0xA0,
    "RSHIFT"             : 0xA1,
    "LCONTROL"           : 0xA2,
    "RCONTROL"           : 0xA3,
    "LMENU"              : 0xA4,
    "RMENU"              : 0xA5,
    "BROWSER_BACK"       : 0xA6,
    "BROWSER_FORWARD"    : 0xA7,
    "BROWSER_REFRESH"    : 0xA8,
    "BROWSER_STOP"       : 0xA9,
    "BROWSER_SEARCH"     : 0xAA,
    "BROWSER_FAVORITES"  : 0xAB,
    "BROWSER_HOME"       : 0xAC,
    "VOLUME_MUTE"        : 0xAD,
    "VOLUME_DOWN"        : 0xAE,
    "VOLUME_UP"          : 0xAF,
    "MEDIA_NEXT_TRACK"   : 0xB0,
    "MEDIA_PREV_TRACK"   : 0xB1,
    "MEDIA_STOP"         : 0xB2,
    "MEDIA_PLAY_PAUSE"   : 0xB3,
    "LAUNCH_MAIL"        : 0xB4,
    "LAUNCH_MEDIA_SELECT": 0xB5,
    "LAUNCH_APP1"        : 0xB6,
    "LAUNCH_APP2"        : 0xB7,
    "OEM_1"              : 0xBA,
    "OEM_PLUS"           : 0xBB,
    "OEM_COMMA"          : 0xBC,
    "OEM_MINUS"          : 0xBD,
    "OEM_PERIOD"         : 0xBE,
    "OEM_2"              : 0xBF,
    "OEM_3"              : 0xC0,
    "OEM_4"              : 0xDB,
    "OEM_5"              : 0xDC,
    "OEM_6"              : 0xDD,
    "OEM_7"              : 0xDE,
    "OEM_8"              : 0xDF,
    "OEM_AX"             : 0xE1,
    "OEM_102"            : 0xE2,
    "ICO_HELP"           : 0xE3,
    "ICO_00"             : 0xE4,
    "PROCESSKEY"         : 0xE5,
    "ICO_CLEAR"          : 0xE6,
    "PACKET"             : 0xE7,
    "OEM_RESET"          : 0xE9,
    "OEM_JUMP"           : 0xEA,
    "OEM_PA1"            : 0xEB,
    "OEM_PA2"            : 0xEC,
    "OEM_PA3"            : 0xED,
    "OEM_WSCTRL"         : 0xEE,
    "OEM_CUSEL"          : 0xEF,
    "OEM_ATTN"           : 0xF0,
    "OEM_FINISH"         : 0xF1,
    "OEM_COPY"           : 0xF2,
    "OEM_AUTO"           : 0xF3,
    "OEM_ENLW"           : 0xF4,
    "OEM_BACKTAB"        : 0xF5,
    "ATTN"               : 0xF6,
    "CRSEL"              : 0xF7,
    "EXSEL"              : 0xF8,
    "EREOF"              : 0xF9,
    "PLAY"               : 0xFA,
    "ZOOM"               : 0xFB,
    "NONAME"             : 0xFC,
    "PA1"                : 0xFD,
    "OEM_CLEAR"          : 0xFE,
}
# fmt: on

vk_dict["HANGEUL"] = vk_dict["KANA"]
vk_dict["HANGUL"] = vk_dict["KANA"]
vk_dict["KANJI"] = vk_dict["HANJA"]
for digit in range(10):
    vk_dict["DIGIT%u" % digit] = 0x30 + digit
for anum in range(26):
    vk_dict[chr(ord("A") + anum)] = 0x41 + anum
VK = enum("VK", vk_dict.items())
VK_TO_NAME = {vk: name for name, vk in vk_dict.items()}

# fmt: off
VK_TO_KEYNAME = {
    VK.ADD                : "kp_add",
    VK.APPS               : "menu",
    VK.BACK               : "backspace",
    VK.BROWSER_BACK       : "back",
    VK.BROWSER_FAVORITES  : "favorites",
    VK.BROWSER_FORWARD    : "forward",
    VK.BROWSER_HOME       : "homepage",
    VK.BROWSER_HOME       : "www",
    VK.BROWSER_REFRESH    : "refresh",
    VK.BROWSER_SEARCH     : "search",
    VK.BROWSER_STOP       : "stop",
    VK.CANCEL             : "cancel",
    VK.CAPITAL            : "caps_lock",
    VK.CLEAR              : "clear",
    VK.DECIMAL            : "kp_decimal",
    VK.DELETE             : "delete",
    VK.DIVIDE             : "kp_divide",
    VK.DOWN               : "down",
    VK.END                : "end",
    VK.ESCAPE             : "escape",
    VK.EXECUTE            : "execute",
    VK.F1                 : "f1",
    VK.F2                 : "f2",
    VK.F3                 : "f3",
    VK.F4                 : "f4",
    VK.F5                 : "f5",
    VK.F6                 : "f6",
    VK.F7                 : "f7",
    VK.F8                 : "f8",
    VK.F9                 : "f9",
    VK.F10                : "f10",
    VK.F11                : "f11",
    VK.F12                : "f12",
    VK.F13                : "f13",
    VK.F14                : "f14",
    VK.F15                : "f15",
    VK.F16                : "f16",
    VK.F17                : "f17",
    VK.F18                : "f18",
    VK.F19                : "f19",
    VK.F20                : "f20",
    VK.F21                : "f21",
    VK.F22                : "f22",
    VK.F23                : "f23",
    VK.F24                : "f24",
    VK.HELP               : "help",
    VK.HOME               : "home",
    VK.INSERT             : "insert",
    VK.LAUNCH_APP1        : "mycomputer",
    VK.LAUNCH_APP2        : "calculator",
    VK.LAUNCH_MAIL        : "mail",
    VK.LAUNCH_MEDIA_SELECT: "audiomedia",
    VK.LCONTROL           : "control_l",
    VK.LEFT               : "left",
    VK.LMENU              : "alt_l",
    VK.LSHIFT             : "shift_l",
    VK.LWIN               : "super_l",
    VK.MEDIA_NEXT_TRACK   : "audionext",
    VK.MEDIA_PLAY_PAUSE   : "audiopause",
    VK.MEDIA_PLAY_PAUSE   : "audioplay",
    VK.MEDIA_PREV_TRACK   : "audioprev",
    VK.MEDIA_STOP         : "audiostop",
    VK.MODECHANGE         : "mode_switch",
    VK.MULTIPLY           : "kp_multiply",
    VK.NEXT               : "page_down",
    VK.NUMLOCK            : "num_lock",
    VK.NUMPAD0            : "kp_0",
    VK.NUMPAD1            : "kp_1",
    VK.NUMPAD2            : "kp_2",
    VK.NUMPAD3            : "kp_3",
    VK.NUMPAD4            : "kp_4",
    VK.NUMPAD5            : "kp_5",
    VK.NUMPAD6            : "kp_6",
    VK.NUMPAD7            : "kp_7",
    VK.NUMPAD8            : "kp_8",
    VK.NUMPAD9            : "kp_9",
    VK.OEM_PLUS           : "kp_equal",
    VK.PAUSE              : "pause",
    VK.PRIOR              : "page_up",
    VK.RCONTROL           : "control_r",
    VK.RETURN             : "return",
    VK.RIGHT              : "right",
    VK.RMENU              : "alt_r",
    VK.RSHIFT             : "shift_r",
    VK.RWIN               : "super_r",
    VK.SCROLL             : "scroll_lock",
    VK.SELECT             : "select",
    VK.SLEEP              : "standby",
    VK.SNAPSHOT           : "print",
    VK.SUBTRACT           : "kp_subtract",
    VK.TAB                : "tab",
    VK.UP                 : "up",
    VK.VOLUME_DOWN        : "audiolowervolume",
    VK.VOLUME_MUTE        : "audiomute",
    VK.VOLUME_UP          : "audioraisevolume",
}
# fmt: on


def vk_to_str(vk):
    s = VK_TO_NAME.get(vk)
    return "%x" % vk if s is None else s


# }}}


DEAD_KEYNAME = {
    "asciicircum": "dead_circumflex",
    "asciitilde": "dead_tilde",
}


class KeyboardLayout:  # {{{
    def __init__(self, layout_id=None, debug=False):
        if layout_id is None:
            layout_id = KeyboardLayout.current_layout_id()
        self.layout_id = layout_id

        # Find virtual key code for each scan code (if any).
        sc_to_vk = {}
        vk_to_sc = {}
        for sc in range(0x01, 0x7F + 1):
            vk = MapVirtualKeyEx(sc, 3, layout_id)
            if vk != 0:
                sc_to_vk[sc] = vk
                vk_to_sc[vk] = sc

        state = (wintypes.BYTE * 256)()
        strbuf = ctypes.create_unicode_buffer(8)

        def fill_state(ss):
            for mod_state, mod_vk in (
                (SHIFT_STATE.SHIFT, VK.SHIFT),
                (SHIFT_STATE.CTRL, VK.CONTROL),
                (SHIFT_STATE.MENU, VK.MENU),
            ):
                state[mod_vk] = 0x80 if (ss & mod_state) != 0 else 0

        def to_unichr(vk, sc, ss):
            fill_state(ss)
            rc = ToUnicodeEx(vk, sc, state, strbuf, len(strbuf), 0, layout_id)
            if rc > 0:
                dead_key = False
                char = ctypes.wstring_at(strbuf.value, rc)
            elif rc < 0:
                # Dead key, flush state.
                dead_key = True
                fill_state(0)
                rc = ToUnicodeEx(
                    VK.SPACE,
                    vk_to_sc[VK.SPACE],
                    state,
                    strbuf,
                    len(strbuf),
                    0,
                    layout_id,
                )
                char = ctypes.wstring_at(strbuf.value, rc) if rc > 0 else ""
            else:
                dead_key = False
                char = ""
            return char, dead_key

        def sort_vk_ss_list(vk_ss_list):
            # Prefer lower modifiers combo.
            return sorted(vk_ss_list, key=lambda vk_ss: popcount_8(vk_ss[1]))

        char_to_vk_ss = defaultdict(list)
        keyname_to_vk = defaultdict(list)

        for sc, vk in sorted(sc_to_vk.items()):
            for ss in SHIFT_STATE:
                if ss in (SHIFT_STATE.MENU, SHIFT_STATE.SHIFT_MENU):
                    # Alt and Shift+Alt don't work, so skip them.
                    continue
                char, dead_key = to_unichr(vk, sc, ss)
                if debug and char:
                    print(
                        "%s%s -> %s [%r] %s"
                        % (
                            shift_state_str(ss),
                            vk_to_str(vk),
                            char,
                            char,
                            "[dead key]" if dead_key else "",
                        )
                    )
                kn = None
                if char:
                    kn = CHAR_TO_KEYNAME.get(char)
                    if dead_key:
                        kn = kn and DEAD_KEYNAME.get(kn, "dead_" + kn)
                    else:
                        char_to_vk_ss[char].append((vk, ss))
                    if kn:
                        keyname_to_vk[kn].append((vk, ss))

        self.char_to_vk_ss = {
            char: sort_vk_ss_list(vk_ss_list)[0]
            for char, vk_ss_list in char_to_vk_ss.items()
        }
        self.char_to_vk_ss["\n"] = self.char_to_vk_ss["\r"]

        self.keyname_to_vk = {kn: vk for vk, kn in VK_TO_KEYNAME.items()}
        self.keyname_to_vk.update(
            {
                "next": VK.NEXT,
                "prior": VK.PRIOR,
                "kp_delete": VK.DELETE,
                "kp_enter": VK.RETURN,
                "break": VK.CANCEL,
            }
        )
        self.keyname_to_vk.update(
            {
                kn: sort_vk_ss_list(vk_ss_list)[0][0]
                for kn, vk_ss_list in keyname_to_vk.items()
            }
        )
        add_modifiers_aliases(self.keyname_to_vk)

        self.ss_to_vks = {}
        for ss in SHIFT_STATE:
            vk_list = []
            for mod_state, mod_vk in (
                (SHIFT_STATE.SHIFT, VK.SHIFT),
                (SHIFT_STATE.CTRL, VK.CONTROL),
                (SHIFT_STATE.MENU, VK.MENU),
            ):
                if (ss & mod_state) != 0:
                    vk_list.append(mod_vk)
            self.ss_to_vks[ss] = tuple(vk_list)

    @staticmethod
    def current_layout_id():
        pid = GetWindowThreadProcessId(GetForegroundWindow(), None)
        return GetKeyboardLayout(pid)


# }}}


if __name__ == "__main__":
    sys.stdout = codecs.getwriter("utf8")(sys.stdout)
    layout = KeyboardLayout(debug=True)
    print("character to virtual key + shift state [%u]" % len(layout.char_to_vk_ss))
    for char, combo in sorted(layout.char_to_vk_ss.items()):
        vk, ss = combo
        print(
            "%s [%r:%s] -> %s%s"
            % (
                char,
                char,
                CHAR_TO_KEYNAME.get(char, "?"),
                shift_state_str(ss),
                vk_to_str(vk),
            )
        )
    print()
    print("keyname to virtual key [%u]" % len(layout.keyname_to_vk))
    for kn, vk in sorted(layout.keyname_to_vk.items()):
        print("%s -> %s" % (kn, vk_to_str(vk)))
    print()
    print("modifiers combo [%u]" % len(layout.ss_to_vks))
    for ss, vk_list in sorted(layout.ss_to_vks.items()):
        print(
            "%s -> %s"
            % (shift_state_str(ss), "+".join(vk_to_str(vk) for vk in vk_list))
        )

# vim: foldmethod=marker



from plover.formatting import Case, SPACE


def meta_mode(ctx, cmdline):
    """
    cmdline should be:

        caps: UPPERCASE
        lower: lowercase
        title: Title Case
        camel: titlecase, no space, initial lowercase
        snake: underscore_space
        reset_space: Space resets to ' '
        reset_case: Reset to normal case
        set_space:xy: Set space to xy
        reset: Reset to normal case, space resets to ' '
    """
    args = cmdline.split(":", 1)
    mode = args.pop(0).lower()
    action = ctx.copy_last_action()
    if mode == "set_space":
        action.space_char = args[0] if args else ""
        return action
    # No argument allowed for other mode directives.
    if args:
        raise ValueError("%r is not a valid mode" % cmdline)
    if mode == "caps":
        action.case = Case.UPPER
    elif mode == "title":
        action.case = Case.TITLE
    elif mode == "lower":
        action.case = Case.LOWER
    elif mode == "snake":
        action.space_char = "_"
    elif mode == "camel":
        action.case = Case.TITLE
        action.space_char = ""
        action.next_case = Case.LOWER_FIRST_CHAR
    elif mode == "reset":
        action.space_char = SPACE
        action.case = None
    elif mode == "reset_space":
        action.space_char = SPACE
    elif mode == "reset_case":
        action.case = None
    else:
        raise ValueError("%r is not a valid mode" % cmdline)
    return action



"""
Thread-based monitoring of a Plover HID stenotype machine.

This implementation is based on the plover-machine-hid plugin created by dnaq (MIT license):
https://github.com/dnaq/plover-machine-hid

Plover HID is a simple HID-based protocol that sends the current state
of the steno machine every time that state changes.
"""

from plover.machine.base import ThreadedStenotypeBase
from plover.misc import boolean
from plover import log

import hid
import time
import platform
import ctypes
import threading
from queue import Queue, Empty
from typing import Dict
from dataclasses import dataclass


def _darwin_disable_exclusive_open():
    lib = getattr(hid, "hidapi", None)
    if lib is not None:
        func = getattr(lib, "hid_darwin_set_open_exclusive")
        func.argtypes = (ctypes.c_int,)
        func.restype = None
        func(0)
    else:
        log.warning(
            "could not call hid_darwin_set_open_exclusive; HID may open exclusively"
        )


# Invoke once at import time (before any devices are opened)
if platform.system() == "Darwin":
    _darwin_disable_exclusive_open()

USAGE_PAGE: int = 0xFF50
USAGE: int = 0x4C56

N_LEVERS: int = 64

# A simple report contains the report id 1 and one bit
# for each of the 64 buttons in the report.
SIMPLE_REPORT_TYPE: int = 0x01
SIMPLE_REPORT_LEN: int = N_LEVERS // 8

# fmt: off
STENO_KEY_CHART = (
    "S1-", "T-", "K-", "P-", "W-", "H-", "R-", "A-",
    "O-", "*1", "-E", "-U", "-F", "-R", "-P", "-B",
    "-L", "-G", "-T", "-S", "-D", "-Z", "#1", "S2-",
    "*2", "*3", "*4", "#2", "#3", "#4", "#5", "#6",
    "#7", "#8", "#9", "#A", "#B", "#C", "X1", "X2",
    "X3", "X4", "X5", "X6", "X7", "X8", "X9", "X10",
    "X11", "X12", "X13", "X14", "X15", "X16", "X17", "X18",
    "X19", "X20", "X21", "X22", "X23", "X24", "X25", "X26",
)
# fmt: on


class InvalidReport(Exception):
    pass


@dataclass
class HidDeviceRecord:
    device: hid.Device
    thread: threading.Thread


class PloverHid(ThreadedStenotypeBase):
    # fmt: off
    KEYS_LAYOUT: str = '''
        #1 #2  #3 #4 #5 #6 #7 #8 #9 #A #B #C
           S1- T- P- H- *1 *3 -F -P -L -T -D
           S2- K- W- R- *2 *4 -R -B -G -S -Z
                  A- O-       -E -U
        X1  X2  X3  X4  X5  X6  X7  X8  X9  X10
        X11 X12 X13 X14 X15 X16 X17 X18 X19 X20
        X21 X22 X23 X24 X25 X26 
    '''
    # fmt: on

    def __init__(self, params):
        super().__init__()
        self._params = params
        self._devices: Dict[str, HidDeviceRecord] = {}
        self._report_queue: Queue[bytes] = Queue()
        self._lock: threading.Lock = threading.Lock()
        self._device_watcher: threading.Thread | None = None

    def _add_device(self, path):
        """Open a HID device at `path`, start its reader, and mark ready if first."""
        try:
            device = hid.Device(path=path)
        except Exception as e:
            log.debug(f"open failed for {path!r}: {e}")
            return False
        was_empty = len(self._devices) == 0
        thread = threading.Thread(
            target=self._read_from_device_loop, args=(path, device), daemon=True
        )
        self._devices[path] = HidDeviceRecord(device=device, thread=thread)
        thread.start()
        if was_empty:
            # We were previously in a disconnected/error state; now we're ready.
            self._ready()
        return True

    def _remove_device(self, path):
        """Close and forget a HID device by path; if none remain, mark disconnected."""
        with self._lock:
            entry = self._devices.pop(path, None)
        if not entry:
            return
        device = entry.device
        thread = entry.thread
        # Closing the device will unblock any pending read in the reader thread.
        try:
            device.close()
        except Exception:
            log.debug("failed to close HID device")
            pass
        # Join the reader if we're not currently in that same thread.
        try:
            if thread is not None and thread is not threading.current_thread():
                thread.join(timeout=0.2)
        except Exception:
            log.debug("failed kill device read thread")
            pass
        # If nothing left and we're not shutting down, show Disconnected in the UI
        if not self._devices and not self.finished.is_set():
            self._error()

    def _scan_device_loop(self):
        """Scan for new matching HID devices and start readers for them."""
        scan_ms = self._params["device_scan_interval_ms"]
        while not self.finished.is_set():
            try:
                paths = [
                    d["path"]
                    for d in hid.enumerate()
                    if d.get("usage_page") == USAGE_PAGE and d.get("usage") == USAGE
                ]
            except Exception as e:
                log.debug(f"device scan enumerate failed: {e}")
                paths = []

            with self._lock:
                for path in paths:
                    if path in self._devices:
                        continue
                    # Call outside lock to avoid holding it during open/start
            for path in paths:
                if path in self._devices:
                    continue
                if self._add_device(path):
                    log.debug("device scan: opened new HID device")

            # sleep but wake early if stopping
            if self.finished.wait(scan_ms / 1000.0):
                break

    def _read_from_device_loop(self, path: str, device: hid.Device) -> None:
        """Per-device reader thread: blocking read, push reports to the report queue."""
        slice_ms = self._params["repeat_interval_ms"]
        while not self.finished.is_set():
            try:
                report = device.read(65536, slice_ms)
            except Exception:
                log.debug("read error: device unplugged?")
                break
            if report:
                try:
                    self._report_queue.put_nowait(report)
                except Exception:
                    log.debug("failed to put report in queue")
                    pass
        self._remove_device(path)

    def _parse(self, report):
        # The first byte is the report id, and due to idiosyncrasies
        # in how HID-apis work on different operating system we can't
        # map the report id to the contents in a good way, so we force
        # compliant devices to always use a report id of 0x50 ('P').
        if len(report) > SIMPLE_REPORT_LEN and report[0] == 0x50:
            return int.from_bytes(report[1 : SIMPLE_REPORT_LEN + 1], "big")
        else:
            raise InvalidReport()

    def _send(self, key_state):
        steno_actions = self.keymap.keys_to_actions(
            [key for i, key in enumerate(STENO_KEY_CHART) if key_state >> (63 - i) & 1]
        )
        if steno_actions:
            self._notify(steno_actions)

    def run(self):
        key_state = 0
        current = 0
        last_sent = 0
        press_started = time.time()
        sent_first_up = False
        while not self.finished.wait(0):
            interval_ms = self._params["repeat_interval_ms"]
            try:
                report = self._report_queue.get(timeout=interval_ms / 1000.0)
            except Empty:
                # No report in this slice; handle repeats
                if (
                    self._params["double_tap_repeat"]
                    and 0 != current == last_sent
                    and time.time() - press_started
                    > self._params["repeat_delay_ms"] / 1e3
                ):
                    self._send(current)
                    # Avoid sending an extra chord when the repeated chord is released.
                    sent_first_up = True
                continue

            try:
                current = self._parse(report)
            except InvalidReport:
                log.error("invalid report")
                continue

            press_started = time.time()
            if self._params["first_up_chord_send"]:
                if key_state & ~current and not sent_first_up:
                    # A finger went up: send a first-up chord and remember it.
                    self._send(key_state)
                    last_sent = key_state
                    sent_first_up = True
                if current & ~key_state:
                    # A finger went down: get ready to send a new first-up chord.
                    sent_first_up = False
                key_state = current
            else:
                key_state |= current
                if current == 0:
                    # All fingers are up: send the "total" chord and reset it.
                    self._send(key_state)
                    last_sent = key_state
                    key_state = 0

    def start_capture(self):
        self.finished.clear()
        self._initializing()
        # Enumerate all hid devices on the machine and if we find one with our
        # usage page and usage we try to connect to it.
        try:
            devices = [
                device["path"]
                for device in hid.enumerate()
                if device["usage_page"] == USAGE_PAGE and device["usage"] == USAGE
            ]

            if not devices:
                log.info("no HID device found; watching for devices…")
                # setting state to error to display Disconnected in the main window
                self._error()

            for path in devices:
                self._add_device(path)
            # Start device watcher
            self._device_watcher = threading.Thread(
                target=self._scan_device_loop, daemon=True
            )
            self._device_watcher.start()
        except Exception as e:
            self._error()
            log.error(f"error during start of capture: {e}")
            return
        self.start()

    def stop_capture(self):
        super().stop_capture()
        # Stop device watcher
        t_watch = getattr(self, "_device_watcher", None)
        if t_watch is not None:
            try:
                t_watch.join(timeout=0.3)
            except Exception:
                pass
            self._device_watcher = None
        # Remove all devices via common teardown
        for path in list(self._devices.keys()):
            try:
                self._remove_device(path)
            except Exception:
                pass
        # Drain the report queue best-effort
        try:
            while True:
                self._report_queue.get_nowait()
        except Exception:
            pass

    @classmethod
    def get_option_info(cls):
        return {
            "first_up_chord_send": (False, boolean),
            "double_tap_repeat": (False, boolean),
            "repeat_delay_ms": (200, int),
            "repeat_interval_ms": (30, int),
            "device_scan_interval_ms": (1000, int),
        }



from functools import partial
import os
import tempfile

import pytest

from plover import system
from plover.config import Config, DictionaryConfig
from plover.engine import ErroredDictionary, StenoEngine
from plover.machine.base import (
    STATE_INITIALIZING,
    STATE_RUNNING,
    STATE_STOPPED,
    StenotypeBase,
)
from plover.machine.keymap import Keymap
from plover.misc import normalize_path
from plover.oslayer.controller import Controller
from plover.output import Output
from plover.registry import Registry
from plover.steno_dictionary import StenoDictionaryCollection

from plover_build_utils.testing import make_dict

from .py37compat import mock


class FakeMachine(StenotypeBase):
    instance = None

    def __init__(self, options):
        super().__init__()
        self.options = options
        self.is_suppressed = False

    @classmethod
    def get_keys(cls):
        return system.KEYS

    def start_capture(self):
        assert FakeMachine.instance is None
        FakeMachine.instance = self
        self._initializing()
        self._ready()

    def stop_capture(self):
        FakeMachine.instance = None
        self._stopped()

    def set_suppression(self, enabled):
        self.is_suppressed = enabled


class FakeKeyboardEmulation(Output):
    def send_backspaces(self, b):
        pass

    def send_string(self, s):
        pass

    def send_key_combination(self, c):
        pass

    def set_key_press_delay(self, delay_ms):
        pass


class FakeEngine(StenoEngine):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.events = []

        def hook_callback(hook, *args, **kwargs):
            self.events.append((hook, args, kwargs))

        for hook in self.HOOKS:
            self.hook_connect(hook, partial(hook_callback, hook))

    def _in_engine_thread(self):
        return True

    def quit(self, code=0):
        self._same_thread_hook(self._quit, code)

    def start(self):
        StenoEngine.start(self)


@pytest.fixture
def engine(monkeypatch):
    FakeMachine.instance = None
    registry = Registry()
    registry.update()
    registry.register_plugin("machine", "Fake", FakeMachine)
    monkeypatch.setattr("plover.config.registry", registry)
    monkeypatch.setattr("plover.engine.registry", registry)
    ctrl = mock.MagicMock(spec=Controller)
    kbd = FakeKeyboardEmulation()
    cfg_file = tempfile.NamedTemporaryFile(
        prefix="plover", suffix="config", delete=False
    )
    try:
        cfg_file.close()
        cfg = Config(cfg_file.name)
        cfg["dictionaries"] = []
        cfg["machine_type"] = "Fake"
        cfg["system_keymap"] = [(k, k) for k in system.KEYS]
        cfg.save()
        yield FakeEngine(cfg, ctrl, kbd)
    finally:
        os.unlink(cfg_file.name)


def test_engine_lifecycle(engine):
    # Config load.
    assert engine.load_config()
    assert engine.events == []
    # Startup.
    engine.start()
    assert engine.events == [
        ("machine_state_changed", ("Fake", "initializing"), {}),
        ("machine_state_changed", ("Fake", "connected"), {}),
        ("config_changed", (engine.config,), {}),
    ]
    assert FakeMachine.instance is not None
    assert not FakeMachine.instance.is_suppressed
    assert len(engine._controller.mock_calls) == 1
    engine._controller.start.assert_called_once()
    engine._controller.reset_mock()
    # Output enabled.
    engine.events.clear()
    engine.output = True
    assert engine.events == [
        ("output_changed", (True,), {}),
    ]
    assert FakeMachine.instance.is_suppressed
    # Machine reconnection.
    engine.events.clear()
    engine.reset_machine()
    assert engine.events == [
        ("machine_state_changed", ("Fake", STATE_STOPPED), {}),
        ("machine_state_changed", ("Fake", STATE_INITIALIZING), {}),
        ("machine_state_changed", ("Fake", STATE_RUNNING), {}),
    ]
    assert FakeMachine.instance is not None
    assert FakeMachine.instance.is_suppressed
    # No machine reset on keymap change.
    engine.events.clear()
    new_keymap = Keymap(system.KEYS, system.KEYS)
    new_keymap.set_mappings(zip(system.KEYS, reversed(system.KEYS)))
    config_update = {"system_keymap": new_keymap}
    assert FakeMachine.instance.keymap != new_keymap
    engine.config = config_update
    assert engine.events == [
        ("config_changed", (config_update,), {}),
    ]
    assert FakeMachine.instance.keymap == new_keymap
    # Output disabled
    engine.events.clear()
    engine.output = False
    assert engine.events == [
        ("output_changed", (False,), {}),
    ]
    assert not FakeMachine.instance.is_suppressed
    # Stopped.
    engine.events.clear()
    engine.quit(42)
    assert engine.join() == 42
    assert engine.events == [
        ("machine_state_changed", ("Fake", STATE_STOPPED), {}),
        ("quit", (), {}),
    ]
    assert FakeMachine.instance is None
    assert len(engine._controller.mock_calls) == 1
    engine._controller.stop.assert_called_once()


def test_loading_dictionaries(tmp_path, engine):
    def check_loaded_events(actual_events, expected_events):
        filtered_events = [
            event for event in actual_events if event[0] != "dictionary_state_changed"
        ]
        assert len(filtered_events) == len(expected_events)
        for n, event in enumerate(filtered_events):
            event_type, event_args, event_kwargs = event
            msg = "event %u: %r" % (n, event)
            assert event_type == "dictionaries_loaded", msg
            assert event_kwargs == {}, msg
            assert len(event_args) == 1, msg
            assert isinstance(event_args[0], StenoDictionaryCollection), msg
            assert [
                (d.path, d.enabled, isinstance(d, ErroredDictionary))
                for d in event_args[0].dicts
            ] == expected_events[n], msg

    with (
        make_dict(tmp_path, b"{}", "json", "valid1") as valid_dict_1,
        make_dict(tmp_path, b"{}", "json", "valid2") as valid_dict_2,
        make_dict(tmp_path, b"", "json", "invalid1") as invalid_dict_1,
        make_dict(tmp_path, b"", "json", "invalid2") as invalid_dict_2,
    ):
        valid_dict_1 = normalize_path(str(valid_dict_1))
        valid_dict_2 = normalize_path(str(valid_dict_2))
        invalid_dict_1 = normalize_path(str(invalid_dict_1))
        invalid_dict_2 = normalize_path(str(invalid_dict_2))
        engine.start()
        for new_dictionaries, *expected_events in (
            # Load one valid dictionary.
            [
                [
                    # path, enabled
                    (valid_dict_1, True),
                ],
                [
                    # path, enabled, errored
                    (valid_dict_1, True, False),
                ],
            ],
            # Load another invalid dictionary.
            [
                [
                    (valid_dict_1, True),
                    (invalid_dict_1, True),
                ],
                [
                    (valid_dict_1, True, False),
                    (invalid_dict_1, True, True),
                ],
            ],
            # Disable first dictionary.
            [
                [
                    (valid_dict_1, False),
                    (invalid_dict_1, True),
                ],
            ],
            # Replace invalid dictionary with another invalid one.
            [
                [
                    (valid_dict_1, False),
                    (invalid_dict_2, True),
                ],
                [
                    (valid_dict_1, False, False),
                ],
                [
                    (valid_dict_1, False, False),
                    (invalid_dict_2, True, True),
                ],
            ],
        ):
            engine.events.clear()
            config_update = {
                "dictionaries": [DictionaryConfig(*d) for d in new_dictionaries]
            }
            engine.config = dict(config_update)
            assert engine.events[0] == ("config_changed", (config_update,), {})
            check_loaded_events(engine.events[1:], expected_events)
        # Simulate an outdated dictionary.
        engine.events.clear()
        engine.dictionaries[valid_dict_1].timestamp -= 1
        engine.config = {}
        check_loaded_events(
            engine.events,
            [
                [
                    (invalid_dict_2, True, True),
                ],
                [
                    (valid_dict_1, False, False),
                    (invalid_dict_2, True, True),
                ],
            ],
        )


def test_engine_running_state(engine):
    # Running state must be different
    # from initial (disabled state).
    initial_state = engine.translator_state
    assert engine.load_config()
    engine.set_output(True)
    running_state = engine.translator_state
    assert running_state != initial_state
    # Disabled state is reset every time
    # output is disabled.
    engine.set_output(False)
    disabled_state = engine.translator_state
    assert disabled_state != running_state
    assert disabled_state != initial_state
    # Running state is kept throughout.
    engine.set_output(True)
    assert engine.translator_state == running_state


def test_undo_and_clear_empty_translator_state(engine):
    engine.clear_translator_state(undo=True)



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

import json
from collections import defaultdict

import pytest

from plover_build_utils.testing import steno_to_stroke


DICT_NAMES = ["main.json", "commands.json", "user.json"]
DICT_PATH = "plover/assets/"


def test_no_duplicates_categorized_files():
    d = defaultdict(list)
    for dictionary in DICT_NAMES:
        with open(DICT_PATH + dictionary, encoding="utf-8") as fp:
            pairs = json.load(fp, object_pairs_hook=lambda x: x)
        for key, value in pairs:
            d[key].append((value, dictionary))
    msg_list = []
    has_duplicate = False
    for key, value_list in d.items():
        if len(value_list) > 1:
            has_duplicate = True
            msg_list.append("key: %s\n" % key)
            for value in value_list:
                msg_list.append("%r in %s\n" % value)
    msg = "\n" + "".join(msg_list)
    assert not has_duplicate, msg


@pytest.mark.parametrize("dictionary", DICT_NAMES)
def test_entries_are_valid(dictionary):
    all_strokes = []
    with open(DICT_PATH + dictionary, encoding="utf-8") as fp:
        for k in json.load(fp):
            all_strokes += k.split("/")
    for s in set(all_strokes):
        steno_to_stroke(s)



import sys

from pygments.formatters import HtmlFormatter
import readme_renderer.markdown
import readme_renderer.rst
import readme_renderer.txt


_RENDERERS = {
    None: readme_renderer.rst,
    "": readme_renderer.rst,
    "text/plain": readme_renderer.txt,
    "text/x-rst": readme_renderer.rst,
    "text/markdown": readme_renderer.markdown,
}

_CSS = "\n".join(
    (
        '<style type="text/css">',
        "pre { background-color: #eeeeee }",
        HtmlFormatter().get_style_defs(),
        "</style>",
    )
)


def description_to_html(content, content_type):
    renderer = _RENDERERS.get(content_type, readme_renderer.rst)
    rendered = renderer.render(content)
    if rendered is None:
        rendered = readme_renderer.txt.render(content)
    return _CSS, rendered


def running_under_virtualenv():
    if sys.prefix != getattr(sys, "base_prefix", sys.prefix):
        # venv
        return True
    if hasattr(sys, "real_prefix"):
        # virtualenv
        return True
    return False



import objc
import logging
from plover import log

NSUserNotification = objc.lookUpClass("NSUserNotification")
NSUserNotificationCenter = objc.lookUpClass("NSUserNotificationCenter")
NSObject = objc.lookUpClass("NSObject")


class NotificationHandler(logging.Handler):
    """Handler using OS X Notification Center to show messages."""

    def __init__(self):
        super().__init__()
        self.setLevel(log.WARNING)
        self.setFormatter(log.NoExceptionTracebackFormatter("%(message)s"))

        self.notification_center = (
            NSUserNotificationCenter.defaultUserNotificationCenter()
        )
        if self.notification_center is None:
            print(
                "no notification center available (e.g. when running from source); not showing notifications"
            )
            self.always_present_delegator = None
        else:
            self.always_present_delegator = AlwaysPresentNSDelegator.alloc().init()

    def handle(self, record):
        if self.notification_center is None:
            # not showing notifications
            return

        # Notification Center has no levels or timeouts.
        notification = NSUserNotification.alloc().init()

        notification.setTitle_(record.levelname.title())
        notification.setInformativeText_(self.format(record))

        self.notification_center.setDelegate_(self.always_present_delegator)
        self.notification_center.deliverNotification_(notification)


class AlwaysPresentNSDelegator(NSObject):
    """
    Custom delegator to force presenting even if Plover is in the foreground.
    """

    def userNotificationCenter_didActivateNotification_(self, ns, note):
        # Do nothing
        return

    def userNotificationCenter_shouldPresentNotification_(self, ns, note):
        # Force notification, even if frontmost application.
        return True



# -*- coding: utf-8 -*-

import plistlib
import os.path

application = defines.get("app", "./dist/Plover.app")
appname = os.path.basename(application)


def icon_from_app(app_path):
    plist_path = os.path.join(app_path, "Contents", "Info.plist")
    with open(plist_path, "rb") as plist_file:
        plist = plistlib.load(plist_file)
    icon_name = plist["CFBundleIconFile"]
    icon_root, icon_ext = os.path.splitext(icon_name)
    if not icon_ext:
        icon_ext = ".icns"
    icon_name = icon_root + icon_ext
    return os.path.join(app_path, "Contents", "Resources", icon_name)


format = defines.get("format", "UDBZ")

# Files to include
files = [application]

# Symlinks to create
symlinks = {"Applications": "/Applications"}

# Volume icon
badge_icon = icon_from_app(application)

# Where to put the icons
icon_locations = {appname: (114, 244), "Applications": (527, 236)}

background = "osx/dmg_resources/background.png"

show_status_bar = False
show_tab_view = False
show_toolbar = False
show_pathbar = False
show_sidebar = False

# Window position in ((x, y), (w, h)) format
window_rect = ((100, 100), (640, 380))

default_view = "icon-view"
show_icon_preview = False
include_icon_view_settings = "auto"
include_list_view_settings = "auto"

arrange_by = None
grid_offset = (0, 0)
grid_spacing = 50
scroll_position = (0, 0)
label_pos = "bottom"
text_size = 16
icon_size = 86



from plover.formatting import Case


def meta_comma(ctx, text):
    action = ctx.new_action()
    action.text = text
    action.prev_attach = True
    return action


def meta_stop(ctx, text):
    action = ctx.new_action()
    action.prev_attach = True
    action.text = text
    action.next_case = Case.CAP_FIRST_WORD
    return action



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""Unit tests for translation.py."""

from collections import namedtuple
import ast
import copy
import operator

from plover.oslayer.config import PLATFORM
from plover.steno import Stroke, normalize_steno

import pytest

from plover.steno_dictionary import StenoDictionary, StenoDictionaryCollection
from plover.translation import Translation, Translator, _State
from plover.translation import escape_translation, unescape_translation

from plover_build_utils.testing import parametrize, steno_to_stroke as stroke


if PLATFORM == "mac":
    BACK_STRING = "{#Alt_L(BackSpace)}{^}"
else:
    BACK_STRING = "{#Control_L(BackSpace)}{^}"


def test_no_translation():
    t = Translation([stroke("S"), stroke("T")], None)
    assert t.strokes == [stroke("S"), stroke("T")]
    assert t.rtfcre == ("S", "T")
    assert t.english is None


def test_translation():
    t = Translation([stroke("S"), stroke("T")], "translation")
    assert t.strokes == [stroke("S"), stroke("T")]
    assert t.rtfcre == ("S", "T")
    assert t.english == "translation"


class TestTranslatorStateSize:
    class FakeState(_State):
        def __init__(self):
            _State.__init__(self)
            self.restrict_calls = []

        def restrict_size(self, n):
            self.restrict_calls.append(n)

    def _check_size_call(self, size):
        assert self.s.restrict_calls[-1] == size

    def _check_no_size_call(self):
        assert self.s.restrict_calls == []

    def clear(self):
        self.s.restrict_calls = []

    def setup_method(self):
        self.t = Translator()
        self.s = type(self).FakeState()
        self.t._state = self.s
        self.d = StenoDictionary()
        self.dc = StenoDictionaryCollection([self.d])
        self.t.set_dictionary(self.dc)

    @pytest.mark.parametrize(
        "key",
        (
            ("S",),
            ("S", "PT", "-Z", "TOP"),
        ),
    )
    def test_dictionary_update_grows_size(self, key):
        self.d[key] = "key"
        self.t.translate(stroke("T-"))
        self._check_size_call(len(key))

    def test_dictionary_update_no_grow(self):
        self.t.set_min_undo_length(4)
        self._check_size_call(4)
        self.clear()
        self.d[("S", "T")] = "nothing"
        self.t.translate(stroke("T-"))
        self._check_size_call(4)

    def test_dictionary_update_shrink(self):
        self.d[("S", "T", "P", "-Z", "-D")] = "1"
        self.t.translate(stroke("T-"))
        self._check_size_call(5)
        self.clear()
        self.d[("A", "P")] = "2"
        self.t.translate(stroke("T-"))
        self._check_size_call(5)
        self.clear()
        del self.d[("S", "T", "P", "-Z", "-D")]
        self.t.translate(stroke("T-"))
        self._check_size_call(2)

    def test_dictionary_update_no_shrink(self):
        self.t.set_min_undo_length(7)
        self.d[("S", "T", "P", "-Z", "-D")] = "1"
        del self.d[("S", "T", "P", "-Z", "-D")]
        self._check_size_call(7)

    def test_translation_calls_restrict(self):
        self.t.translate(stroke("S"))
        self._check_size_call(0)


def test_listeners():
    output1 = []

    def listener1(undo, do, prev):
        output1.append((undo, do, prev))

    output2 = []

    def listener2(undo, do, prev):
        output2.append((undo, do, prev))

    t = Translator()
    s = stroke("S")
    tr = Translation([s], None)
    expected_output = [([], [tr], [tr])]

    t.translate(s)

    t.add_listener(listener1)
    t.translate(s)
    assert output1 == expected_output

    del output1[:]
    t.add_listener(listener2)
    t.translate(s)
    assert output1 == expected_output
    assert output2 == expected_output

    del output1[:]
    del output2[:]
    t.add_listener(listener2)
    t.translate(s)
    assert output1 == expected_output
    assert output2 == expected_output

    del output1[:]
    del output2[:]
    t.remove_listener(listener1)
    t.translate(s)
    assert output1 == []
    assert output2 == expected_output

    del output1[:]
    del output2[:]
    t.remove_listener(listener2)
    t.translate(s)
    assert output1 == []
    assert output2 == []


def test_changing_state():
    output = []

    def listener(undo, do, prev):
        prev = list(prev) if prev else None
        output.append((undo, do, prev))

    d = StenoDictionary()
    d[("S", "P")] = "hi"
    dc = StenoDictionaryCollection([d])
    t = Translator()
    t.set_dictionary(dc)
    t.translate(stroke("T"))
    t.translate(stroke("S"))
    s = copy.deepcopy(t.get_state())

    t.add_listener(listener)

    expected = [
        (
            [Translation([stroke("S")], None)],
            [Translation([stroke("S"), stroke("P")], "hi")],
            [Translation([stroke("T")], None)],
        )
    ]
    t.translate(stroke("P"))
    assert output == expected

    del output[:]
    t.set_state(s)
    t.translate(stroke("P"))
    assert output == expected

    del output[:]
    t.clear_state()
    t.translate(stroke("P"))
    assert output == [([], [Translation([stroke("P")], None)], None)]

    del output[:]
    t.set_state(s)
    t.translate(stroke("P"))
    assert output == [
        (
            [],
            [Translation([stroke("P")], None)],
            [Translation([stroke("S"), stroke("P")], "hi")],
        )
    ]


def test_translator():
    # It's not clear that this test is needed anymore. There are separate
    # tests for _translate_stroke and test_translate_calls_translate_stroke
    # makes sure that translate calls it properly. But since I already wrote
    # this test I'm going to keep it.

    class Output:
        def __init__(self):
            self._output = []

        def write(self, undo, do, prev):
            for t in undo:
                self._output.pop()
            for t in do:
                if t.english:
                    self._output.append(t.english)
                else:
                    self._output.append("/".join(t.rtfcre))

        def get(self):
            return " ".join(self._output)

        def clear(self):
            del self._output[:]

    d = StenoDictionary()
    out = Output()
    t = Translator()
    dc = StenoDictionaryCollection([d])
    t.set_dictionary(dc)
    t.add_listener(out.write)

    t.translate(stroke("S"))
    assert out.get() == "S"
    t.translate(stroke("T"))
    assert out.get() == "S T"
    t.translate(stroke("*"))
    assert out.get() == "S"
    t.translate(stroke("*"))
    # Undo buffer ran out
    assert out.get() == "S " + BACK_STRING

    t.set_min_undo_length(3)
    out.clear()
    t.translate(stroke("S"))
    assert out.get() == "S"
    t.translate(stroke("T"))
    assert out.get() == "S T"
    t.translate(stroke("*"))
    assert out.get() == "S"
    t.translate(stroke("*"))
    assert out.get() == ""

    out.clear()
    d[("S",)] = "t1"
    d[("T",)] = "t2"
    d[("S", "T")] = "t3"

    t.translate(stroke("S"))
    assert out.get() == "t1"
    t.translate(stroke("T"))
    assert out.get() == "t3"
    t.translate(stroke("T"))
    assert out.get() == "t3 t2"
    t.translate(stroke("S"))
    assert out.get() == "t3 t2 t1"
    t.translate(stroke("*"))
    assert out.get() == "t3 t2"
    t.translate(stroke("*"))
    assert out.get() == "t3"
    t.translate(stroke("*"))
    assert out.get() == "t1"
    t.translate(stroke("*"))
    assert out.get() == ""

    t.translate(stroke("S"))
    assert out.get() == "t1"
    t.translate(stroke("T"))
    assert out.get() == "t3"
    t.translate(stroke("T"))
    assert out.get() == "t3 t2"

    d[("S", "T", "T")] = "t4"
    d[("S", "T", "T", "S")] = "t5"

    t.translate(stroke("S"))
    assert out.get() == "t5"
    t.translate(stroke("*"))
    assert out.get() == "t3 t2"
    t.translate(stroke("*"))
    assert out.get() == "t3"
    t.translate(stroke("T"))
    assert out.get() == "t4"
    t.translate(stroke("S"))
    assert out.get() == "t5"
    t.translate(stroke("S"))
    assert out.get() == "t5 t1"
    t.translate(stroke("*"))
    assert out.get() == "t5"
    t.translate(stroke("*"))
    assert out.get() == "t4"
    t.translate(stroke("*"))
    assert out.get() == "t3"
    t.translate(stroke("*"))
    assert out.get() == "t1"
    t.translate(stroke("*"))
    assert out.get() == ""

    d.clear()

    s = stroke("S")
    t.translate(s)
    t.translate(s)
    t.translate(s)
    t.translate(s)
    s = stroke("*")
    t.translate(s)
    t.translate(s)
    t.translate(s)
    t.translate(s)
    # Not enough undo to clear output.
    assert out.get() == "S " + BACK_STRING

    out.clear()
    t.remove_listener(out.write)
    t.translate(stroke("S"))
    assert out.get() == ""


class TestState:
    def setup_method(self):
        self.a = Translation([stroke("S")], None)
        self.b = Translation([stroke("T"), stroke("-D")], None)
        self.c = Translation([stroke("-Z"), stroke("P"), stroke("T*")], None)

    def test_prev_list0(self):
        s = _State()
        assert s.prev() is None

    def test_prev_list1(self):
        s = _State()
        s.translations = [self.a]
        assert s.prev() == [self.a]

    def test_prev_list2(self):
        s = _State()
        s.translations = [self.a, self.b]
        assert s.prev() == [self.a, self.b]

    def test_prev_tail1(self):
        s = _State()
        s.translations = [self.a]
        s.tail = self.b
        assert s.prev() == [self.a]

    def test_prev_tail0(self):
        s = _State()
        s.tail = self.b
        assert s.prev() == [self.b]

    def test_restrict_size_zero_on_empty(self):
        s = _State()
        s.restrict_size(0)
        assert s.translations == []
        assert s.tail is None

    def test_restrict_size_zero_on_one_stroke(self):
        s = _State()
        s.translations = [self.a]
        s.restrict_size(0)
        assert s.translations == [self.a]
        assert s.tail is None

    def test_restrict_size_to_exactly_one_stroke(self):
        s = _State()
        s.translations = [self.a]
        s.restrict_size(1)
        assert s.translations == [self.a]
        assert s.tail is None

    def test_restrict_size_to_one_on_two_strokes(self):
        s = _State()
        s.translations = [self.b]
        s.restrict_size(1)
        assert s.translations == [self.b]
        assert s.tail is None

    def test_restrict_size_to_one_on_two_translations(self):
        s = _State()
        s.translations = [self.b, self.a]
        s.restrict_size(1)
        assert s.translations == [self.a]
        assert s.tail == self.b

    def test_restrict_size_to_one_on_two_translations_too_big(self):
        s = _State()
        s.translations = [self.a, self.b]
        s.restrict_size(1)
        assert s.translations == [self.b]
        assert s.tail == self.a

    def test_restrict_size_lose_translations(self):
        s = _State()
        s.translations = [self.a, self.b, self.c]
        s.restrict_size(2)
        assert s.translations == [self.c]
        assert s.tail == self.b

    def test_restrict_size_multiple_translations(self):
        s = _State()
        s.translations = [self.a, self.b, self.c]
        s.restrict_size(5)
        assert s.translations == [self.b, self.c]
        assert s.tail == self.a


class TestTranslateStroke:
    DICT_COLLECTION_CLASS = StenoDictionaryCollection

    class CaptureOutput:
        Output = namedtuple("Output", "undo do prev")

        def __init__(self):
            self.output = []

        def __call__(self, undo, new, prev):
            prev = list(prev) if prev else None
            self.output = self.Output(undo, new, prev)

    def t(self, strokes):
        """A quick way to make a translation."""
        strokes = [stroke(x) for x in strokes.split("/")]
        key = tuple(s.rtfcre for s in strokes)
        translation = self.dc.lookup(key)
        return Translation(strokes, translation)

    def lt(self, translations):
        """A quick way to make a list of translations."""
        return [self.t(x) for x in translations.split()]

    def define(self, key, value):
        key = normalize_steno(key)
        self.d[key] = value

    def translate(self, steno):
        self.tlor.translate(stroke(steno))

    def _check_translations(self, expected):
        # Hide from traceback on assertions (reduce output size for failed tests).
        __tracebackhide__ = operator.methodcaller("errisinstance", AssertionError)
        msg = """
        translations:
            results: %s
            expected: %s
        """ % (self.s.translations, expected)
        assert self.s.translations == expected, msg

    def _check_output(self, undo, do, prev):
        # Hide from traceback on assertions (reduce output size for failed tests).
        __tracebackhide__ = operator.methodcaller("errisinstance", AssertionError)
        msg = """
        output:
            results: -%s
                     +%s
                     [%s]
            expected: -%s
                      +%s
                      [%s]
        """ % (self.o.output + (undo, do, prev))
        assert self.o.output == (undo, do, prev), msg

    def setup_method(self):
        self.d = StenoDictionary()
        self.dc = self.DICT_COLLECTION_CLASS([self.d])
        self.s = _State()
        self.o = self.CaptureOutput()
        self.tlor = Translator()
        self.tlor.set_dictionary(self.dc)
        self.tlor.add_listener(self.o)
        self.tlor.set_state(self.s)

    def test_first_stroke(self):
        self.translate("-B")
        self._check_translations(self.lt("-B"))
        self._check_output([], self.lt("-B"), None)

    def test_second_stroke(self):
        self.define("S/P", "spiders")
        self.s.translations = self.lt("S")
        self.translate("-T")
        self._check_translations(self.lt("S -T"))
        self._check_output([], self.lt("-T"), self.lt("S"))

    def test_second_stroke_tail(self):
        self.s.tail = self.t("T/A/EU/L")
        self.translate("-E")
        self._check_translations(self.lt("E"))
        self._check_output([], self.lt("E"), self.lt("T/A/EU/L"))

    def test_with_translation_1(self):
        self.define("S", "is")
        self.define("-T", "that")
        self.s.translations = self.lt("S")
        self.tlor.set_min_undo_length(2)
        self.translate("-T")
        self._check_translations(self.lt("S -T"))
        self._check_output([], self.lt("-T"), self.lt("S"))
        assert self.o.output.do[0].english == "that"

    def test_with_translation_2(self):
        self.define("S", "is")
        self.define("-T", "that")
        self.s.translations = self.lt("S")
        self.tlor.set_min_undo_length(1)
        self.translate("-T")
        self._check_translations(self.lt("-T"))
        self._check_output([], self.lt("-T"), self.lt("S"))
        assert self.o.output.do[0].english == "that"

    def test_finish_two_translation(self):
        self.define("S/T", "hello")
        self.s.translations = self.lt("S")
        self.translate("T")
        self._check_translations(self.lt("S/T"))
        self._check_output(self.lt("S"), self.lt("S/T"), None)
        assert self.o.output.do[0].english == "hello"
        assert self.o.output.do[0].replaced == self.lt("S")

    def test_finish_three_translation(self):
        self.define("S/T/-B", "bye")
        self.s.translations = self.lt("S T")
        self.translate("-B")
        self._check_translations(self.lt("S/T/-B"))
        self._check_output(self.lt("S T"), self.lt("S/T/-B"), None)
        assert self.o.output.do[0].english == "bye"
        assert self.o.output.do[0].replaced == self.lt("S T")

    def test_replace_translation(self):
        self.define("S/T/-B", "longer")
        self.s.translations = self.lt("S/T")
        self.translate("-B")
        self._check_translations(self.lt("S/T/-B"))
        self._check_output(self.lt("S/T"), self.lt("S/T/-B"), None)
        assert self.o.output.do[0].english == "longer"
        assert self.o.output.do[0].replaced == self.lt("S/T")

    def test_undo(self):
        self.s.translations = self.lt("POP")
        self.translate("*")
        self._check_translations([])
        self._check_output(self.lt("POP"), [], None)

    def test_empty_undo(self):
        self.translate("*")
        self._check_translations([])
        self._check_output([], [Translation([Stroke("*")], BACK_STRING)], None)

    def test_undo_translation(self):
        self.define("P/P", "pop")
        self.translate("P")
        self.translate("P")
        self.translate("*")
        self._check_translations(self.lt("P"))
        self._check_output(self.lt("P/P"), self.lt("P"), None)

    def test_undo_longer_translation(self):
        self.define("P/P/-D", "popped")
        self.translate("P")
        self.translate("P")
        self.translate("-D")
        self.translate("*")
        self._check_translations(self.lt("P P"))
        self._check_output(self.lt("P/P/-D"), self.lt("P P"), None)

    def test_undo_tail(self):
        self.s.tail = self.t("T/A/EU/L")
        self.translate("*")
        self._check_translations([])
        self._check_output([], [Translation([Stroke("*")], BACK_STRING)], [self.s.tail])

    def test_suffix_folding(self):
        self.define("K-L", "look")
        self.define("-G", "{^ing}")
        lt = self.lt("K-LG")
        lt[0].english = "look {^ing}"
        self.translate("K-LG")
        self._check_translations(lt)

    def test_suffix_folding_multi_stroke(self):
        self.define("E/HR", "he will")
        self.define("-S", "{^s}")
        self.translate("E")
        self.translate("HR-S")
        output = " ".join(t.english for t in self.s.translations)
        assert output == "he will {^s}"

    def test_suffix_folding_doesnt_interfere(self):
        self.define("E/HR", "he will")
        self.define("-S", "{^s}")
        self.define("E", "he")
        self.define("HR-S", "also")
        self.translate("E")
        self.translate("HR-S")
        output = " ".join(t.english for t in self.s.translations)
        assert output == "he also"

    def test_suffix_folding_no_suffix(self):
        self.define("K-L", "look")
        lt = self.lt("K-LG")
        assert lt[0].english is None
        self.translate("K-LG")
        self._check_translations(lt)

    def test_suffix_folding_no_main(self):
        self.define("-G", "{^ing}")
        lt = self.lt("K-LG")
        assert lt[0].english is None
        self.translate("K-LG")
        self._check_translations(lt)

    def test_retrospective_insert_space(self):
        self.define("T/E/S/T", "a longer key")
        self.define("PER", "perfect")
        self.define("SWAEUGS", "situation")
        self.define("PER/SWAEUGS", "persuasion")
        self.define("SP*", "{*?}")
        self.translate("PER")
        self.translate("SWAEUGS")
        self.translate("SP*")
        lt = self.lt("PER")
        undo = self.lt("PER/SWAEUGS")
        undo[0].replaced = lt
        do = self.lt("SP*")
        do[0].english = "perfect situation"
        do[0].is_retrospective_command = True
        do[0].replaced = undo
        self._check_translations(do)
        self._check_output(undo, do, None)

    def test_retrospective_insert_space_undefined(self):
        # Should work when beginning or ending strokes aren't defined
        self.define("T/E/S/T", "a longer key")
        self.define("STWR/STWR", "test")
        self.define("SP*", "{*?}")
        self.translate("STWR")
        self.translate("STWR")
        self.translate("SP*")
        lt = self.lt("STWR")
        undo = self.lt("STWR/STWR")
        undo[0].replaced = lt
        do = self.lt("SP*")
        do[0].english = "STWR STWR"
        do[0].is_retrospective_command = True
        do[0].replaced = undo
        self._check_translations(do)
        self._check_output(undo, do, None)

    def test_retrospective_delete_space(self):
        self.define("T/E/S/T", "a longer key")
        self.define("K", "kick")
        self.define("PW", "back")
        self.define("SP*", "{*!}")
        self.translate("K")
        self.translate("PW")
        self.translate("SP*")
        undo = self.lt("K PW")
        do = self.lt("SP*")
        do[0].english = "kick{^~|^}back"
        do[0].is_retrospective_command = True
        do[0].replaced = undo
        self._check_translations(do)
        self._check_output(undo, do, None)

    def test_retrospective_delete_space_with_number(self):
        self.define("T/E/S/T", "a longer key")
        self.define("U", "user")
        self.define("SP*", "{*!}")
        self.translate("U")
        self.translate("1-")
        self.translate("SP*")
        undo = self.lt("U 1-")
        do = self.lt("SP*")
        do[0].english = "user{^~|^}{&1}"
        do[0].is_retrospective_command = True
        do[0].replaced = undo
        self._check_translations(do)
        self._check_output(undo, do, None)

    def test_retrospective_delete_space_with_period(self):
        self.define("T/E/S/T", "a longer key")
        self.define("P-P", "{.}")
        self.define("SH*", "zshrc")
        self.define("SP*", "{*!}")
        self.translate("P-P")
        self.translate("SH*")
        self.translate("SP*")
        undo = self.lt("P-P SH*")
        do = self.lt("SP*")
        do[0].english = "{.}{^~|^}zshrc"
        do[0].is_retrospective_command = True
        do[0].replaced = undo
        self._check_translations(do)
        self._check_output(undo, do, None)

    def test_retrospective_toggle_asterisk(self):
        self.define("T/E/S/T", "a longer key")
        self.define("S", "see")
        self.define("S*", "sea")
        self.define("A*", "{*}")
        self.translate("S")
        self.translate("A*")
        self._check_translations(self.lt("S*"))
        self._check_output(self.lt("S"), self.lt("S*"), None)

    def test_retrospective_toggle_empty(self):
        self.define("A*", "{*}")
        self.translate("A*")
        self._check_translations(self.lt(""))
        assert self.o.output == []

    def test_retrospective_toggle_asterisk_replaced1(self):
        self.define("P-P", "{.}")
        self.define("SKEL", "cancel")
        self.define("SKEL/TO-PB", "skeleton")
        self.define("SKEL/TO*PB", "not skeleton!")
        self.define("A*", "{*}")
        self.translate("P-P")
        self.translate("SKEL")
        self.translate("TO-PB")
        self.translate("A*")
        self._check_translations(self.lt("SKEL/TO*PB"))
        self._check_output(self.lt("SKEL/TO-PB"), self.lt("SKEL/TO*PB"), self.lt("P-P"))

    def test_retrospective_toggle_asterisk_replaced2(self):
        self.define("P-P", "{.}")
        self.define("SKEL", "cancel")
        self.define("SKEL/TO-PB", "skeleton")
        self.define("TO*PB", "{^ton}")
        self.define("A*", "{*}")
        self.translate("P-P")
        self.translate("SKEL")
        self.translate("TO-PB")
        self.translate("A*")
        self._check_translations(self.lt("SKEL TO*PB"))
        self._check_output(self.lt("SKEL/TO-PB"), self.lt("SKEL TO*PB"), self.lt("P-P"))

    def test_repeat_last_stroke1(self):
        self.define("T/E/S/T", "a longer key")
        self.define("TH", "this")
        self.define("R*", "{*+}")
        self.translate("TH")
        self.translate("R*")
        undo = []
        do = self.lt("TH")
        state = self.lt("TH TH")
        self._check_translations(state)
        self._check_output(undo, do, do)

    def test_repeat_last_stroke2(self):
        self.define("T/E/S/T", "a longer key")
        self.define("THA", "that")
        self.define("R*", "{*+}")
        self.translate("THA")
        self.translate("R*")
        undo = []
        do = self.lt("THA")
        state = self.lt("THA THA")
        self._check_translations(state)
        self._check_output(undo, do, do)

    def test_untranslate_translation(self):
        self.tlor.set_min_undo_length(2)
        self.define("TH", "this")
        self.define("THA", "that")
        self.translate("TH")
        self.translate("THA")
        self.tlor.untranslate_translation(self.t("THA"))
        self.tlor.untranslate_translation(self.t("TH"))
        self.tlor.flush()
        self._check_output(self.lt("TH THA"), [], None)


ESCAPE_UNESCAPE_TRANSLATION_TESTS = (
    # No change.
    lambda: ("foobar", "foobar"),
    lambda: (r"\\", r"\\"),
    lambda: ("\\\\\\", "\\\\\\"),  # -> \\\
    # Basic support: \n, \r, \t.
    lambda: ("\n", r"\n"),
    lambda: ("\r", r"\r"),
    lambda: ("\t", r"\t"),
    # Allow a literal \n, \r, or \t by doubling the \.
    lambda: (r"\n", r"\\n"),
    lambda: (r"\r", r"\\r"),
    lambda: (r"\t", r"\\t"),
    lambda: (r"\\n", r"\\\n"),
    lambda: (r"\\r", r"\\\r"),
    lambda: (r"\\t", r"\\\t"),
    # A little more complex.
    lambda: ("\tfoo\nbar\r", r"\tfoo\nbar\r"),
    lambda: ("\\tfoo\\nbar\\r", r"\\tfoo\\nbar\\r"),
)


@parametrize(ESCAPE_UNESCAPE_TRANSLATION_TESTS)
def test_escape_unescape_translation(raw, escaped):
    assert unescape_translation(escaped) == raw
    assert escape_translation(raw) == escaped


class TestNoUnnecessaryLookups(TestTranslateStroke):
    # Custom dictionary collection class for tracking lookups.
    class DictTracy(StenoDictionaryCollection):
        def __init__(self, dicts):
            super().__init__(dicts)
            self.lookup_history = []

        def lookup(self, key):
            self.lookup_history.append(key)
            return super().lookup(key)

    DICT_COLLECTION_CLASS = DictTracy

    def _prepare_state(self, definitions, translations):
        if definitions:
            for steno, english in ast.literal_eval("{" + definitions + "}").items():
                self.define(steno, english)
        translations = self.lt(translations)
        for t in translations:
            for s in t.strokes:
                self.translate(s.rtfcre)
        state = translations[len(translations) - self.dc.longest_key :]
        self._check_translations(state)
        self.dc.lookup_history.clear()

    def _check_lookup_history(self, expected):
        # Hide from traceback on assertions (reduce output size for failed tests).
        __tracebackhide__ = operator.methodcaller("errisinstance", AssertionError)
        result = ["/".join(key) for key in self.dc.lookup_history]
        expected = expected.split()
        msg = """
        lookup history:
            results: %s
            expected: %s
        """ % (result, expected)
        assert result == expected, msg

    def test_zero_lookups(self):
        # No lookups at all if longest key is zero.
        self.translate("TEFT")
        self._check_lookup_history("")
        self._check_translations(self.lt("TEFT"))

    def test_no_prefix_lookup_over_the_longest_key_limit(self):
        self._prepare_state(
            """
            "HROPBG/EFT/KAOE": "longest key",
            "HRETS": "let's",
            "TKO": "do",
            "SPH": "some",
            "TEFT": "test",
            "-G": "{^ing}",
            """,
            "HRETS TKO SPH TEFT",
        )
        self.translate("-G")
        self._check_lookup_history(
            # Macros.
            """
            /-G
            -G
            """
            # Others.
            """
            SPH/TEFT/-G
            /TEFT/-G TEFT/-G
            """
        )

    def test_no_duplicate_lookups_for_longest_no_suffix_match(self):
        self._prepare_state(
            """
            "TEFT": "test",
            "-G": "{^ing}",
            """,
            "TEFT",
        )
        self.translate("TEFGT")
        self._check_lookup_history(
            # Macros.
            """
            TEFGT
            """
            # No suffix.
            """
            """
            # With suffix.
            """
            -G TEFT
            """
        )

    def test_lookup_suffixes_once(self):
        self._prepare_state(
            """
            "HROPBG/EFT/KAOE": "longest key",
            "HRETS": "let's",
            "TEFT": "test",
            "SPH": "some",
            "SUFBGS": "suffix",
            "-G": "{^ing}",
            "-S": "{^s}",
            "-D": "{^ed}",
            "-Z": "{^s}",
            """,
            "HRETS TEFT SPH",
        )
        self.translate("SUFBGSZ")
        self._check_lookup_history(
            # Macros.
            """
            /SUFBGSZ
            SUFBGSZ
            """
            # Without suffix.
            """
            TEFT/SPH/SUFBGSZ
            /SPH/SUFBGSZ
            SPH/SUFBGSZ
            """
            # Suffix lookups.
            """
            -Z -S -G
            TEFT/SPH/SUFBGS TEFT/SPH/SUFBGZ TEFT/SPH/SUFBSZ
            /SPH/SUFBGS /SPH/SUFBGZ /SPH/SUFBSZ
            SPH/SUFBGS SPH/SUFBGZ SPH/SUFBSZ
            /SUFBGS /SUFBGZ /SUFBSZ
            SUFBGS
            """
        )



from concurrent.futures import as_completed
import json
import os

from plover.plugins_manager.requests import CachedFuturesSession


PYPI_URL = "https://pypi.org/pypi"
REGISTRY_URL = "https://raw.githubusercontent.com/openstenoproject/plover_plugins_registry/master/registry.json"


def find_plover_plugins_releases(pypi_url=None, registry_url=None, capture=None):
    if pypi_url is None:
        pypi_url = os.environ.get("PYPI_URL", PYPI_URL)

    if registry_url is None:
        registry_url = os.environ.get("REGISTRY_URL", REGISTRY_URL)

    session = CachedFuturesSession()

    in_progress = set()
    all_releases = {}

    def fetch_release(name, version=None):
        if (name, version) in all_releases:
            return
        all_releases[(name, version)] = None
        if version is None:
            url = "%s/%s/json" % (pypi_url, name)
        else:
            url = "%s/%s/%s/json" % (pypi_url, name, version)
        in_progress.add(session.get(url))

    with session:
        for name in session.get(registry_url).result().json():
            fetch_release(name)

        while in_progress:
            for future in as_completed(list(in_progress)):
                in_progress.remove(future)
                if not future.done():
                    continue
                resp = future.result()
                if resp.status_code != 200:
                    # Can happen if a package has been deleted.
                    continue
                release = resp.json()
                info = release["info"]
                if "plover_plugin" not in info["keywords"].split():
                    # Not a plugin.
                    continue
                name, version = info["name"], info["version"]
                all_releases[(name, version)] = release

    all_releases = [release for release in all_releases.values() if release is not None]

    if capture is not None:
        with open(capture, "w") as fp:
            json.dump(all_releases, fp, indent=2, sort_keys=True)

    return all_releases



# Copyright (c) 2011 Hesky Fisher.
# See LICENSE.txt for details.

"""This package abstracts os details for plover."""

import os

from plover import log

from .config import PLATFORM


PLATFORM_PACKAGE = {
    "bsd": "linux",
    "linux": "linux",
    "mac": "osx",
    "win": "windows",
}


def _add_platform_package_to_path():
    platform_package = PLATFORM_PACKAGE.get(PLATFORM)
    if platform_package is None:
        log.warning("No platform-specific oslayer package for: %s" % PLATFORM)
        return
    __path__.insert(0, os.path.join(__path__[0], platform_package))


_add_platform_package_to_path()



#!/usr/bin/env python3

import os
import sys
import subprocess


def get_pip(args=None):
    os.makedirs("build", exist_ok=True)

    # can be downloaded from https://github.com/pypa/get-pip/tags by following the commit ID and opening the raw file
    cmd = [
        "curl",
        "https://raw.githubusercontent.com/pypa/get-pip/c8048075753de535e6279e501940f41bc040c081/public/get-pip.py",
        "-o",
        "build/get-pip.py",
    ]
    subprocess.call(cmd)
    cmd = [sys.executable, "build/get-pip.py", "pip==25.1"]
    subprocess.call(cmd)
    if args:
        cmd = [sys.executable, "-m", "pip", "install"] + args
        print("running", " ".join(cmd), flush=True)
        subprocess.check_call(cmd)


if __name__ == "__main__":
    get_pip(sys.argv[1:])



def meta_retro_currency(ctx, dict_format):
    action = ctx.copy_last_action()
    last_words = ctx.last_words(count=1)
    if not last_words:
        return action
    currency = last_words[0].replace(",", "")
    for cast, fmt in (
        (int, "{:,}"),
        (float, "{:,.2f}"),
    ):
        try:
            cast_input = cast(currency)
        except ValueError:
            continue
        currency_format = dict_format.replace("c", fmt)
        action.prev_attach = True
        action.prev_replace = last_words[0]
        action.text = currency_format.format(cast_input)
        action.word = None
        break
    return action



from plover import log, __version__

from plover.plugins_manager import global_registry, local_registry

from plover.plugins_manager.requests import CachedFuturesSession


class PackageState:
    def __init__(self, name, installed=None, available=None):
        self.name = name
        self.installed = installed or []
        self.available = available or []
        self.status = "installed" if installed else ""

    @property
    def current(self):
        return self.installed[-1] if self.installed else None

    @current.setter
    def current(self, metadata):
        self.installed.append(metadata)
        self.status = "removed" if metadata is None else "updated"

    @property
    def latest(self):
        return self.available[-1] if self.available else None

    @property
    def metadata(self):
        return self.current or self.latest

    def __getattr__(self, name):
        return getattr(self.metadata, name)

    def __str__(self):
        s = self.name
        if self.latest:
            s += " " + self.latest.version
        if self.current:
            s += " [" + self.status + " " + self.current.version + "]"
        return s

    def __lt__(self, other):
        return self.name < other.name

    def __repr__(self):
        return str(self)


UNSUPPORTED_PLUGINS_URL = "https://raw.githubusercontent.com/openstenoproject/plover_plugins_registry/master/unsupported.json"


class Registry:
    def __init__(self):
        self._packages = {
            name: PackageState(name, installed=metadata)
            for name, metadata in local_registry.list_plugins().items()
        }

    def __len__(self):
        return len(self._packages)

    def __contains__(self, name):
        return name in self._packages

    def __getitem__(self, name):
        return self._packages[name]

    def __iter__(self):
        return iter(self._packages.values())

    def keys(self):
        return self._packages.keys()

    def items(self):
        return self._packages.items()

    # TODO add tests for this method
    def parse_unsupported_plover_version(self, unsupported_plover_version) -> int:
        """
        Handling different formats in case the unsupported_plover_version format changes in future Plover versions.
        """
        if isinstance(unsupported_plover_version, int):
            return unsupported_plover_version
        elif isinstance(unsupported_plover_version, str):
            try:
                return int(unsupported_plover_version.split(".")[0])
            except Exception as e:
                raise ValueError(
                    f'Failed to parse unsupported plover version "{unsupported_plover_version}" from plugin metadata'
                ) from e
        else:
            raise ValueError(
                f'Unknown format for unsupported plover version "{unsupported_plover_version}" from plugin metadata'
            )

    def is_plugin_supported(self, pkg, unsupported_plugins_dict):
        if not unsupported_plugins_dict:
            return True
        if pkg.name in unsupported_plugins_dict:
            unsupported_plover_version = unsupported_plugins_dict[pkg.name]
            try:
                parsed_unsupported_plover_version = (
                    self.parse_unsupported_plover_version(unsupported_plover_version)
                )
            except:
                log.warning(
                    f'Failed to parse unsupported plover version "{pkg.unsupported_plover_version}" for plugin {pkg.name}, assuming plugin is supported',
                    exc_info=True,
                )
                return True
            current_major_plover_version = int(__version__.split(".")[0])
            return current_major_plover_version < parsed_unsupported_plover_version
        else:
            return True

    def update(self):
        try:
            available_plugins = global_registry.list_plugins()
        except:
            log.error(
                "Failed to fetch list of available plugins from PyPI", exc_info=True
            )
            return

        session = CachedFuturesSession()
        try:
            unsupported_plugins_dict = (
                session.get(UNSUPPORTED_PLUGINS_URL).result().json()
            )
        except:
            log.warning(
                "Failed to fetch list of unsupported plugins, assuming all plugins are supported",
                exc_info=True,
            )

        for name, metadata in available_plugins.items():
            pkg = self._packages.get(name)
            if pkg is None:
                pkg = PackageState(name, available=metadata)
                self._packages[name] = pkg
            else:
                pkg.available = metadata
                if (
                    pkg.current
                    and pkg.current.parsed_version < pkg.latest.parsed_version
                ):
                    pkg.status = "outdated"
            if pkg.status != "installed" and not self.is_plugin_supported(
                pkg, unsupported_plugins_dict
            ):
                pkg.status = "unsupported"



from PySide6.QtCore import QObject, Signal
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QMessageBox, QSystemTrayIcon

from plover import _, __name__ as __software_name__
from plover import log
from plover.oslayer.config import PLATFORM
from plover.machine.base import (
    STATE_STOPPED,
    STATE_INITIALIZING,
    STATE_RUNNING,
    STATE_ERROR,
)


class TrayIcon(QObject):
    def __init__(self):
        super().__init__()
        self._supported = QSystemTrayIcon.isSystemTrayAvailable()
        self._context_menu = None
        self._enabled = False
        self._trayicon = None
        self._state_icons = {}
        for state in (
            "disconnected",
            "disabled",
            "enabled",
        ):
            icon = QIcon(":resources/state-%s.svg" % state)
            if hasattr(icon, "setIsMask"):
                icon.setIsMask(True)
            self._state_icons[state] = icon
        self._machine = None
        self._machine_state = "disconnected"
        self._is_running = False
        self._update_state()

    def set_menu(self, menu):
        self._context_menu = menu
        if self._enabled:
            self._trayicon.setContextMenu(menu)

    def show_message(
        self, message, icon=QSystemTrayIcon.MessageIcon.Information, timeout=10000
    ):
        self._trayicon.showMessage(
            __software_name__.capitalize(), message, icon, timeout
        )

    def log(self, level, message):
        if self._enabled:
            if level <= log.INFO:
                icon = QSystemTrayIcon.MessageIcon.Information
                timeout = 10
            elif level <= log.WARNING:
                icon = QSystemTrayIcon.MessageIcon.Warning
                timeout = 15
            else:
                icon = QSystemTrayIcon.MessageIcon.Critical
                timeout = 25
            self.show_message(message, icon, timeout * 1000)
        else:
            if level <= log.INFO:
                icon = QMessageBox.Information
            elif level <= log.WARNING:
                icon = QMessageBox.Warning
            else:
                icon = QMessageBox.Critical
            msgbox = QMessageBox()
            msgbox.setText(message)
            msgbox.setIcon(icon)
            msgbox.exec()

    def is_supported(self):
        return self._supported

    def enable(self):
        if not self._supported:
            return
        self._trayicon = QSystemTrayIcon()
        # On OS X, the context menu is activated with either mouse buttons,
        # and activation messages are still sent, so ignore those...
        if PLATFORM != "mac":
            self._trayicon.activated.connect(self._on_activated)
        if self._context_menu is not None:
            self._trayicon.setContextMenu(self._context_menu)
        self._enabled = True
        self._update_state()
        self._trayicon.show()

    def disable(self):
        if not self._enabled:
            return
        self._trayicon.hide()
        self._trayicon = None
        self._enabled = False

    def is_enabled(self):
        return self._enabled

    def update_machine_state(self, machine, state):
        self._machine = machine
        self._machine_state = state
        self._update_state()

    def update_output(self, enabled):
        self._is_running = enabled
        self._update_state()

    clicked = Signal()

    def _update_state(self):
        if self._machine_state not in (STATE_INITIALIZING, STATE_RUNNING):
            state = "disconnected"
        else:
            state = "enabled" if self._is_running else "disabled"
        icon = self._state_icons[state]
        if not self._enabled:
            return
        machine_state = _("{machine} is {state}").format(
            machine=_(self._machine),
            state=_(self._machine_state),
        )
        if self._is_running:
            # i18n: Tray icon tooltip.
            output_state = _("output is enabled")
        else:
            # i18n: Tray icon tooltip.
            output_state = _("output is disabled")
        self._trayicon.setIcon(icon)
        self._trayicon.setToolTip(
            # i18n: Tray icon tooltip.
            "Plover:\n- %s\n- %s" % (output_state, machine_state)
        )

    def _on_activated(self, reason):
        if reason == QSystemTrayIcon.ActivationReason.Trigger:
            self.clicked.emit()



from PySide6.QtWidgets import QDialog

from plover.gui_qt.utils import WindowStateMixin


class Tool(QDialog, WindowStateMixin):
    # Used for dialog window title, menu entry text.
    TITLE = None
    # Optional path to icon image.
    ICON = None
    # Optional shortcut (as a string).
    SHORTCUT = None
    # Note: the class documentation is automatically used as tooltip.

    def __init__(self, engine):
        super().__init__()
        self._update_title()
        self._engine = engine

    def _update_title(self):
        self.setWindowTitle("Plover: " + self.TITLE)

    def setupUi(self, widget):
        super().setupUi(widget)
        self._update_title()



from collections import namedtuple
from html import escape as html_escape
from os.path import split as os_path_split

from PySide6.QtCore import QEvent, Signal, Slot
from PySide6.QtGui import QFontDatabase
from PySide6.QtWidgets import QApplication, QWidget

from plover import _
from plover.misc import shorten_path
from plover.steno import normalize_steno, sort_steno_strokes
from plover.engine import StartingStrokeState
from plover.translation import escape_translation, unescape_translation
from plover.formatting import RetroFormatter
from plover.resource import resource_filename

from plover.gui_qt.add_translation_widget_ui import Ui_AddTranslationWidget
from plover.gui_qt.steno_validator import StenoValidator


class AddTranslationWidget(QWidget, Ui_AddTranslationWidget):
    # i18n: Widget: “AddTranslationWidget”, tooltip.
    __doc__ = _("Add a new translation to the dictionary.")

    EngineState = namedtuple(
        "EngineState", "dictionary_filter translator starting_stroke"
    )

    mappingValid = Signal(bool)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setupUi(self)
        engine = QApplication.instance().engine
        self._engine = engine
        self._dictionaries = []
        self._reverse_order = False
        self._selected_dictionary = None
        self._mapping_is_valid = False
        engine.signal_connect("config_changed", self.on_config_changed)
        self.on_config_changed(engine.config)
        engine.signal_connect("dictionaries_loaded", self.on_dictionaries_loaded)
        self.on_dictionaries_loaded(self._engine.dictionaries)

        fixed_family = QFontDatabase.systemFont(
            QFontDatabase.SystemFont.FixedFont
        ).family()

        self._special_fmt = f'<span style="font-family:{fixed_family};">%s</span>'
        self._special_fmt_bold = (
            f'<span style="font-family:{fixed_family};font-weight:bold;">%s</span>'
        )

        self.strokes.setValidator(StenoValidator())
        self.strokes.installEventFilter(self)
        self.translation.installEventFilter(self)

        with engine:
            # Pre-populate the strokes or translations with last stroke/word.
            last_translations = engine.translator_state.translations
            translation = None
            for t in reversed(last_translations):
                # Find the last undoable stroke.
                if t.has_undo():
                    translation = t
                    break
            # Is it a raw stroke?
            if translation is not None and not translation.english:
                # Yes.
                self.strokes.setText(translation.formatting[0].text)
                self.handle_stroke_input_change()
                self.strokes.selectAll()
            else:
                # No, grab the last-formatted word.
                retro_formatter = RetroFormatter(last_translations)
                last_words = retro_formatter.last_words(strip=True)
                if last_words:
                    self.translation.setText(last_words[0])
                    self.handle_translation_input_change()

            self._original_state = self.EngineState(
                None, engine.translator_state, engine.starting_stroke_state
            )
            engine.clear_translator_state()
            self._strokes_state = self.EngineState(
                self._dictionary_filter,
                engine.translator_state,
                StartingStrokeState(True, False, "/"),
            )
            engine.clear_translator_state()
            self._translations_state = self.EngineState(
                None, engine.translator_state, StartingStrokeState(True, False, " ")
            )
        self._engine_state = self._original_state
        self._focus = None

    @property
    def mapping_is_valid(self):
        return self._mapping_is_valid

    def select_dictionary(self, dictionary_path):
        self._selected_dictionary = dictionary_path
        self._update_items()

    def eventFilter(self, watched, event):
        if event.type() == QEvent.Type.FocusIn:
            if watched == self.strokes:
                self._focus_strokes()
            elif watched == self.translation:
                self._focus_translation()
        elif event.type() == QEvent.Type.FocusOut:
            if watched in (self.strokes, self.translation):
                self._unfocus()
        return False

    def _set_engine_state(self, state):
        with self._engine as engine:
            prev_state = self._engine_state
            if prev_state is not None and prev_state.dictionary_filter is not None:
                engine.remove_dictionary_filter(prev_state.dictionary_filter)
            engine.translator_state = state.translator
            engine.starting_stroke_state = state.starting_stroke
            if state.dictionary_filter is not None:
                engine.add_dictionary_filter(state.dictionary_filter)
            self._engine_state = state

    @staticmethod
    def _dictionary_filter(key, value):
        # Allow undo...
        if value == "=undo":
            return False
        # ...and translations with special entries. Do this by looking for
        # braces but take into account escaped braces and slashes.
        escaped = value.replace("\\\\", "").replace("\\{", "")
        special = "{#" in escaped or "{PLOVER:" in escaped
        return not special

    def _unfocus(self):
        self._unfocus_strokes()
        self._unfocus_translation()

    def _focus_strokes(self):
        if self._focus == "strokes":
            return
        self._unfocus_translation()
        self._set_engine_state(self._strokes_state)
        self._focus = "strokes"

    def _unfocus_strokes(self):
        if self._focus != "strokes":
            return
        self._set_engine_state(self._original_state)
        self._focus = None

    def _focus_translation(self):
        if self._focus == "translation":
            return
        self._unfocus_strokes()
        self._set_engine_state(self._translations_state)
        self._focus = "translation"

    def _unfocus_translation(self):
        if self._focus != "translation":
            return
        self._set_engine_state(self._original_state)
        self._focus = None

    def _strokes(self):
        strokes = self.strokes.text().strip()
        has_prefix = strokes.startswith("/")
        strokes = "/".join(strokes.replace("/", " ").split())
        if has_prefix:
            strokes = "/" + strokes
        strokes = normalize_steno(strokes)
        return strokes

    def _translation(self):
        translation = self.translation.text().strip()
        return unescape_translation(translation)

    def _update_items(self, dictionaries=None, reverse_order=None):
        if dictionaries is not None:
            self._dictionaries = dictionaries
        if reverse_order is not None:
            self._reverse_order = reverse_order
        iterable = self._dictionaries
        if self._reverse_order:
            iterable = reversed(iterable)
        self.dictionary.clear()
        for d in iterable:
            item = shorten_path(d.path)
            if not d.enabled:
                # i18n: Widget: “AddTranslationWidget”.
                item = _("{dictionary} (disabled)").format(dictionary=item)
            self.dictionary.addItem(item)
        selected_index = 0
        if self._selected_dictionary is None:
            # No user selection, select first enabled dictionary.
            for n, d in enumerate(self._dictionaries):
                if d.enabled:
                    selected_index = n
                    break
        else:
            # Keep user selection.
            for n, d in enumerate(self._dictionaries):
                if d.path == self._selected_dictionary:
                    selected_index = n
                    break
        if self._reverse_order:
            selected_index = self.dictionary.count() - selected_index - 1
        self.dictionary.setCurrentIndex(selected_index)

    def on_dictionaries_loaded(self, dictionaries):
        # We only care about loaded writable dictionaries.
        dictionaries = [d for d in dictionaries.dicts if not d.readonly]
        if dictionaries != self._dictionaries:
            self._update_items(dictionaries=dictionaries)

    def on_config_changed(self, config_update):
        if "classic_dictionaries_display_order" in config_update:
            self._update_items(
                reverse_order=config_update["classic_dictionaries_display_order"]
            )

    @Slot(int)
    def update_selected_dictionary(self, index):
        if self._reverse_order:
            index = len(self._dictionaries) - index - 1
        self._selected_dictionary = self._dictionaries[index].path

    def _format_label(self, fmt, strokes, translation=None, filename=None):
        if strokes:
            strokes = ", ".join(
                self._special_fmt % html_escape("/".join(s))
                for s in sort_steno_strokes(strokes)
            )
        if translation:
            translation = self._special_fmt_bold % html_escape(
                escape_translation(translation)
            )

        if filename:
            filename = html_escape(filename)

        return fmt.format(strokes=strokes, translation=translation, filename=filename)

    @Slot()
    def handle_stroke_input_change(self):
        mapping_is_valid = self.strokes.hasAcceptableInput()
        if mapping_is_valid != self._mapping_is_valid:
            self._mapping_is_valid = mapping_is_valid
            self.mappingValid.emit(mapping_is_valid)
        if not mapping_is_valid:
            return
        strokes = self._strokes()
        if strokes:
            translations = self._engine.raw_lookup_from_all(strokes)
            if translations:
                # i18n: Widget: “AddTranslationWidget”.
                info = self._format_label(_("{strokes} maps to "), (strokes,))
                entries = [
                    self._format_label(
                        ("• " if i else "") + "<bf>{translation}<bf/>\t({filename})",
                        None,
                        translation,
                        os_path_split(resource_filename(dictionary.path))[1],
                    )
                    for i, (translation, dictionary) in enumerate(translations)
                ]
                if len(entries) > 1:
                    # i18n: Widget: “AddTranslationWidget”.
                    entries.insert(1, "<br />" + _("Overwritten entries:"))
                info += "<br />".join(entries)
            else:
                info = self._format_label(
                    # i18n: Widget: “AddTranslationWidget”.
                    _("{strokes} is not mapped in any dictionary"),
                    (strokes,),
                )
        else:
            info = ""
        self.strokes_info.setText(info)

    @Slot()
    def handle_translation_input_change(self):
        translation = self._translation()
        if translation:
            strokes = self._engine.reverse_lookup(translation)
            if strokes:
                # i18n: Widget: “AddTranslationWidget”.
                fmt = _("{translation} is mapped to: {strokes}")
            else:
                # i18n: Widget: “AddTranslationWidget”.
                fmt = _("{translation} is not in the dictionary")
            info = self._format_label(fmt, strokes, translation)
        else:
            info = ""
        self.translation_info.setText(info)

    def save_entry(self):
        self._unfocus()
        strokes = self._strokes()
        translation = self._translation()
        if strokes and translation:
            index = self.dictionary.currentIndex()
            if self._reverse_order:
                index = -index - 1
            dictionary = self._dictionaries[index]
            old_translation = self._engine.dictionaries[dictionary.path].get(strokes)
            self._engine.add_translation(
                strokes, translation, dictionary_path=dictionary.path
            )
            return dictionary, strokes, old_translation, translation

    def reject(self):
        self._unfocus()
        self._set_engine_state(self._original_state)



from pathlib import Path
import signal
import sys

from PySide6.QtCore import (
    QCoreApplication,
    QLibraryInfo,
    QTimer,
    QTranslator,
    QtMsgType,
    qInstallMessageHandler,
)
from PySide6.QtWidgets import QApplication, QMessageBox

from plover import _, __name__ as __software_name__, __version__, log
from plover.oslayer.config import CONFIG_DIR
from plover.oslayer.keyboardcontrol import KeyboardEmulation

from plover.gui_qt.engine import Engine


# Disable input hook to avoid getting spammed when using the debugger.
# import pdb
# pdb.set_trace()


class Application:
    def __init__(self, config, controller, use_qt_notifications):
        # This is done dynamically so localization
        # support can be configure beforehand.
        from plover.gui_qt.main_window import MainWindow

        self._app = None
        self._win = None
        self._engine = None
        self._translator = None

        QCoreApplication.setApplicationName(__software_name__.capitalize())
        QCoreApplication.setApplicationVersion(__version__)
        QCoreApplication.setOrganizationName("Open Steno Project")
        QCoreApplication.setOrganizationDomain("openstenoproject.org")

        self._app = QApplication([sys.argv[0], "-name", "plover"])

        # Apply custom stylesheet if present.
        stylesheet_path = Path(CONFIG_DIR) / "plover.qss"
        if stylesheet_path.exists():
            log.info("using stylesheet at: %s", stylesheet_path)
            self._app.setStyleSheet(stylesheet_path.read_text())

        # Enable localization of standard Qt controls.
        log.info("setting language to: %s", _.lang)
        self._translator = QTranslator()
        translations_dir = QLibraryInfo.path(QLibraryInfo.LibraryPath.TranslationsPath)
        self._translator.load("qtbase_" + _.lang, translations_dir)
        self._app.installTranslator(self._translator)

        QApplication.setQuitOnLastWindowClosed(False)

        self._app.engine = self._engine = Engine(
            config, controller, KeyboardEmulation()
        )
        # On macOS, quitting through the dock will result
        # in a direct call to `QCoreApplication.quit`.
        self._app.aboutToQuit.connect(self._app.engine.quit)

        signal.signal(signal.SIGINT, lambda signum, stack: self._engine.quit())

        # Make sure the Python interpreter runs at least every second,
        # so signals have a chance to be processed.
        self._timer = QTimer()
        self._timer.timeout.connect(lambda: None)
        self._timer.start(1000)

        self._win = MainWindow(self._engine, use_qt_notifications)

    def __del__(self):
        del self._win
        del self._app
        del self._engine
        del self._translator

    def run(self):
        self._app.exec()
        return self._engine.join()


def show_error(title, message):
    print("%s: %s" % (title, message))
    app = QApplication([])
    QMessageBox.critical(None, title, message)
    del app


def default_excepthook(*exc_info):
    log.error("Qt GUI error", exc_info=exc_info)


def default_message_handler(msg_type, msg_log_context, msg_string):
    log_fn = {
        QtMsgType.QtDebugMsg: log.debug,
        QtMsgType.QtInfoMsg: log.info,
        QtMsgType.QtWarningMsg: log.warning,
    }.get(msg_type, log.error)
    details = []
    if msg_log_context.file is not None:
        details.append("%s:%u" % (msg_log_context.file, msg_log_context.line))
    if msg_log_context.function is not None:
        details.append(msg_log_context.function)
    if details:
        details = " [%s]" % ", ".join(details)
    else:
        details = ""
    log_fn("Qt: %s%s", msg_string, details)


def main(config, controller):
    # Setup internationalization support.
    use_qt_notifications = not log.has_platform_handler()
    # Log GUI exceptions that make it back to the event loop.
    if sys.excepthook is sys.__excepthook__:
        sys.excepthook = default_excepthook
    # Capture and log Qt messages.
    qInstallMessageHandler(default_message_handler)
    app = Application(config, controller, use_qt_notifications)
    code = app.run()
    del app
    return code
